# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module contains abstract classes for the various clients of the Azure IoT Hub Device SDK
"""
from __future__ import annotations  # Needed for annotation bug < 3.10
import abc
import logging
import threading
import os
import io
import time
from . import pipeline
from .pipeline import constant as pipeline_constant
from azure.iot.device.common.auth import connection_string as cs
from azure.iot.device.common.auth import sastoken as st
from azure.iot.device.iothub import client_event
from azure.iot.device.iothub.models import Message, MethodRequest, MethodResponse
from azure.iot.device.common.models import X509
from azure.iot.device import exceptions
from azure.iot.device.common import auth, handle_exceptions
from . import edge_hsm
from .pipeline import MQTTPipeline, HTTPPipeline
from typing_extensions import Self
from azure.iot.device.custom_typing import FunctionOrCoroutine, Twin, TwinPatch
from typing import Any, Dict, List, Optional, Union

logger = logging.getLogger(__name__)


def _validate_kwargs(exclude: Optional[List[str]] = [], **kwargs) -> None:
    """Helper function to validate user provided kwargs.
    Raises TypeError if an invalid option has been provided"""
    valid_kwargs = [
        "server_verification_cert",
        "gateway_hostname",
        "websockets",
        "cipher",
        "product_info",
        "proxy_options",
        "sastoken_ttl",
        "keep_alive",
        "auto_connect",
        "connection_retry",
        "connection_retry_interval",
        "ensure_desired_properties",
    ]

    for kwarg in kwargs:
        if (kwarg not in valid_kwargs) or (exclude is not None and kwarg in exclude):
            raise TypeError("Unsupported keyword argument: '{}'".format(kwarg))


def _get_config_kwargs(**kwargs) -> Dict[str, Any]:
    """Get the subset of kwargs which pertain the config object"""
    valid_config_kwargs = [
        "server_verification_cert",
        "gateway_hostname",
        "websockets",
        "cipher",
        "product_info",
        "proxy_options",
        "keep_alive",
        "auto_connect",
        "connection_retry",
        "connection_retry_interval",
        "ensure_desired_properties",
    ]

    config_kwargs = {}
    for kwarg in kwargs:
        if kwarg in valid_config_kwargs:
            config_kwargs[kwarg] = kwargs[kwarg]
    return config_kwargs


def _form_sas_uri(hostname: str, device_id: str, module_id: Optional[str] = None) -> str:
    if module_id:
        return "{hostname}/devices/{device_id}/modules/{module_id}".format(
            hostname=hostname, device_id=device_id, module_id=module_id
        )
    else:
        return "{hostname}/devices/{device_id}".format(hostname=hostname, device_id=device_id)


def _extract_sas_uri_values(uri: str) -> Dict[str, Any]:
    d = {}
    items = uri.split("/")
    if len(items) != 3 and len(items) != 5:
        raise ValueError("Invalid SAS URI")
    if items[1] != "devices":
        raise ValueError("Cannot extract device id from SAS URI")
    if len(items) > 3 and items[3] != "modules":
        raise ValueError("Cannot extract module id from SAS URI")
    d["hostname"] = items[0]
    d["device_id"] = items[2]
    try:
        d["module_id"] = items[4]
    except IndexError:
        d["module_id"] = ""
    return d


# Receive Type constant defs
RECEIVE_TYPE_NONE_SET = "none_set"  # Type of receiving has not been set
RECEIVE_TYPE_HANDLER = "handler"  # Only use handlers for receive
RECEIVE_TYPE_API = "api"  # Only use APIs for receive


class AbstractIoTHubClient(abc.ABC):
    """A superclass representing a generic IoTHub client.
    This class needs to be extended for specific clients.
    """

    def __init__(self, mqtt_pipeline: MQTTPipeline, http_pipeline: HTTPPipeline) -> None:
        """Initializer for a generic client.

        :param mqtt_pipeline: The pipeline used to connect to the IoTHub endpoint.
        :type mqtt_pipeline: :class:`azure.iot.device.iothub.pipeline.MQTTPipeline`
        """
        self._mqtt_pipeline = mqtt_pipeline
        self._http_pipeline = http_pipeline

        self._inbox_manager = None  # this will be overridden in child class
        self._handler_manager = None  # this will be overridden in child class
        self._receive_type = RECEIVE_TYPE_NONE_SET
        self._client_lock = threading.Lock()

    def _on_connected(self) -> None:
        """Helper handler that is called upon an iothub pipeline connect"""
        logger.info("Connection State - Connected")
        if self._inbox_manager is not None:
            client_event_inbox = self._inbox_manager.get_client_event_inbox()
            # Only add a ClientEvent to the inbox if the Handler Manager is capable of dealing with it
            if self._handler_manager.handling_client_events:
                event = client_event.ClientEvent(client_event.CONNECTION_STATE_CHANGE)
                client_event_inbox.put(event)
            # Ensure that all handlers are running now that connection is re-established.
            self._handler_manager.ensure_running()

    def _on_disconnected(self) -> None:
        """Helper handler that is called upon an iothub pipeline disconnect"""
        logger.info("Connection State - Disconnected")
        if self._inbox_manager is not None:
            client_event_inbox = self._inbox_manager.get_client_event_inbox()
            # Only add a ClientEvent to the inbox if the Handler Manager is capable of dealing with it
            if self._handler_manager.handling_client_events:
                event = client_event.ClientEvent(client_event.CONNECTION_STATE_CHANGE)
                client_event_inbox.put(event)
            # Locally stored method requests on client are cleared.
            # They will be resent by IoTHub on reconnect.
            self._inbox_manager.clear_all_method_requests()
            logger.info("Cleared all pending method requests due to disconnect")

    def _on_new_sastoken_required(self) -> None:
        """Helper handler that is called upon the iothub pipeline needing new SAS token"""
        logger.info("New SasToken required from user")
        if self._inbox_manager is not None:
            client_event_inbox = self._inbox_manager.get_client_event_inbox()
            # Only add a ClientEvent to the inbox if the Handler Manager is capable of dealing with it
            if self._handler_manager.handling_client_events:
                event = client_event.ClientEvent(client_event.NEW_SASTOKEN_REQUIRED)
                client_event_inbox.put(event)

    def _on_background_exception(self, e: Exception) -> None:
        """Helper handler that is called upon an iothub pipeline background exception"""
        handle_exceptions.handle_background_exception(e)
        if self._inbox_manager is not None:
            client_event_inbox = self._inbox_manager.get_client_event_inbox()
            # Only add a ClientEvent to the inbox if the Handler Manager is capable of dealing with it
            if self._handler_manager.handling_client_events:
                event = client_event.ClientEvent(client_event.BACKGROUND_EXCEPTION, e)
                client_event_inbox.put(event)

    def _check_receive_mode_is_api(self) -> None:
        """Call this function first in EVERY receive API"""
        with self._client_lock:
            if self._receive_type is RECEIVE_TYPE_NONE_SET:
                # Lock the client to ONLY use receive APIs (no handlers)
                self._receive_type = RECEIVE_TYPE_API
            elif self._receive_type is RECEIVE_TYPE_HANDLER:
                raise exceptions.ClientError(
                    "Cannot use receive APIs - receive handler(s) have already been set"
                )
            else:
                pass

    def _check_receive_mode_is_handler(self) -> None:
        """Call this function first in EVERY handler setter"""
        with self._client_lock:
            if self._receive_type is RECEIVE_TYPE_NONE_SET:
                # Lock the client to ONLY use receive handlers (no APIs)
                self._receive_type = RECEIVE_TYPE_HANDLER
                # Set the inbox manager to use unified msg receives
                if self._inbox_manager is not None:
                    self._inbox_manager.use_unified_msg_mode = True
            elif self._receive_type is RECEIVE_TYPE_API:
                raise exceptions.ClientError(
                    "Cannot set receive handlers - receive APIs have already been used"
                )
            else:
                pass

    def _replace_user_supplied_sastoken(self, sastoken_str: str) -> None:
        """
        Replaces the pipeline's NonRenewableSasToken with a new one based on a provided
        sastoken string. Also does validation.
        This helper only updates the PipelineConfig - it does not reauthorize the connection.
        """
        if not isinstance(
            self._mqtt_pipeline.pipeline_configuration.sastoken, st.NonRenewableSasToken
        ):
            raise exceptions.ClientError(
                "Cannot update sastoken when client was not created with one"
            )
        # Create new SasToken
        try:
            new_token_o = st.NonRenewableSasToken(sastoken_str)
        except st.SasTokenError as e:
            new_err = ValueError("Invalid SasToken provided")
            new_err.__cause__ = e
            raise new_err
        # Extract values from SasToken
        vals = _extract_sas_uri_values(new_token_o.resource_uri)
        # Validate new token
        if type(self).__name__ == "IoTHubDeviceClient" and vals["module_id"]:
            raise ValueError("Provided SasToken is for a module")
        if type(self).__name__ == "IoTHubModuleClient" and not vals["module_id"]:
            raise ValueError("Provided SasToken is for a device")
        if self._mqtt_pipeline.pipeline_configuration.device_id != vals["device_id"]:
            raise ValueError("Provided SasToken does not match existing device id")
        if (
            vals["module_id"] != ""
            and self._mqtt_pipeline.pipeline_configuration.module_id != vals["module_id"]
        ):
            raise ValueError("Provided SasToken does not match existing module id")
        if self._mqtt_pipeline.pipeline_configuration.hostname != vals["hostname"]:
            raise ValueError("Provided SasToken does not match existing hostname")
        if new_token_o.expiry_time < int(time.time()):
            raise ValueError("Provided SasToken has already expired")
        # Set token
        # NOTE: We only need to set this on MQTT because this is a reference to the same object
        # that is stored in HTTP. The HTTP pipeline is updated implicitly.
        self._mqtt_pipeline.pipeline_configuration.sastoken = new_token_o

    @abc.abstractmethod
    def _generic_receive_handler_setter(
        self,
        handler_name: str,
        feature_name: str,
        new_handler: Optional[FunctionOrCoroutine[[Any], Any]],
    ) -> None:
        # Will be implemented differently in child classes, but define here for static analysis
        pass

    @classmethod
    def create_from_connection_string(cls, connection_string: str, **kwargs) -> Self:
        """
        Instantiate the client from a IoTHub device or module connection string.

        :param str connection_string: The connection string for the IoTHub you wish to connect to.

        :param str server_verification_cert: Configuration Option. The trusted certificate chain.
            Necessary when using connecting to an endpoint which has a non-standard root of trust,
            such as a protocol gateway.
        :param bool websockets: Configuration Option. Default is False. Set to true if using MQTT
            over websockets.
        :param cipher: Configuration Option. Cipher suite(s) for TLS/SSL, as a string in
            "OpenSSL cipher list format" or as a list of cipher suite strings.
        :type cipher: str or list(str)
        :param str product_info: Configuration Option. Default is empty string. The string contains
            arbitrary product info which is appended to the user agent string.
        :param proxy_options: Options for sending traffic through proxy servers.
        :type proxy_options: :class:`azure.iot.device.ProxyOptions`
        :param int sastoken_ttl: The time to live (in seconds) for the created SasToken used for
            authentication. Default is 3600 seconds (1 hour).
        :param int keep_alive: Maximum period in seconds between communications with the
            broker. If no other messages are being exchanged, this controls the
            rate at which the client will send ping messages to the broker.
            If not provided default value of 60 secs will be used.
        :param bool auto_connect: Automatically connect the client to IoTHub when a method is
            invoked which requires a connection to be established. (Default: True)
        :param bool connection_retry: Attempt to re-establish a dropped connection (Default: True)
        :param int connection_retry_interval: Interval, in seconds, between attempts to
            re-establish a dropped connection (Default: 10)
        :param bool ensure_desired_properties: Ensure the most recent desired properties patch has
            been received upon re-connections (Default:True)

        :raises: ValueError if given an invalid connection_string.
        :raises: TypeError if given an unsupported parameter.

        :returns: An instance of an IoTHub client that uses a connection string for authentication.
        """
        # TODO: Make this device/module specific and reject non-matching connection strings.

        # Ensure no invalid kwargs were passed by the user
        excluded_kwargs = ["gateway_hostname"]
        _validate_kwargs(exclude=excluded_kwargs, **kwargs)

        # Create SasToken
        connection_string_dict = cs.ConnectionString(connection_string)
        if connection_string_dict.get(cs.X509) is not None:
            raise ValueError(
                "Use the .create_from_x509_certificate() method instead when using X509 certificates"
            )
        uri = _form_sas_uri(
            hostname=connection_string_dict[cs.HOST_NAME],
            device_id=connection_string_dict[cs.DEVICE_ID],
            module_id=connection_string_dict.get(cs.MODULE_ID),
        )
        signing_mechanism = auth.SymmetricKeySigningMechanism(
            key=connection_string_dict[cs.SHARED_ACCESS_KEY]
        )
        token_ttl = kwargs.get("sastoken_ttl", 3600)
        try:
            sastoken = st.RenewableSasToken(uri, signing_mechanism, ttl=token_ttl)
        except st.SasTokenError as e:
            new_err = ValueError("Could not create a SasToken using provided values")
            new_err.__cause__ = e
            raise new_err
        # Pipeline Config setup
        config_kwargs = _get_config_kwargs(**kwargs)
        pipeline_configuration = pipeline.IoTHubPipelineConfig(
            device_id=connection_string_dict[cs.DEVICE_ID],
            module_id=connection_string_dict.get(cs.MODULE_ID),
            hostname=connection_string_dict[cs.HOST_NAME],
            gateway_hostname=connection_string_dict.get(cs.GATEWAY_HOST_NAME),
            sastoken=sastoken,
            **config_kwargs,
        )
        if cls.__name__ == "IoTHubDeviceClient":
            pipeline_configuration.blob_upload = True

        # Pipeline setup
        http_pipeline = pipeline.HTTPPipeline(pipeline_configuration)
        mqtt_pipeline = pipeline.MQTTPipeline(pipeline_configuration)

        return cls(mqtt_pipeline, http_pipeline)

    @classmethod
    def create_from_sastoken(cls, sastoken: str, **kwargs: Dict[str, Any]) -> Self:
        """Instantiate the client from a pre-created SAS Token string

        :param str sastoken: The SAS Token string

        :param str server_verification_cert: Configuration Option. The trusted certificate chain.
            Necessary when using connecting to an endpoint which has a non-standard root of trust,
            such as a protocol gateway.
        :param str gateway_hostname: Configuration Option. The gateway hostname for the gateway
            device.
        :param bool websockets: Configuration Option. Default is False. Set to true if using MQTT
            over websockets.
        :param cipher: Configuration Option. Cipher suite(s) for TLS/SSL, as a string in
            "OpenSSL cipher list format" or as a list of cipher suite strings.
        :type cipher: str or list(str)
        :param str product_info: Configuration Option. Default is empty string. The string contains
            arbitrary product info which is appended to the user agent string.
        :param proxy_options: Options for sending traffic through proxy servers.
        :type proxy_options: :class:`azure.iot.device.ProxyOptions`
        :param int keep_alive: Maximum period in seconds between communications with the
            broker. If no other messages are being exchanged, this controls the
            rate at which the client will send ping messages to the broker.
            If not provided default value of 60 secs will be used.
        :param bool auto_connect: Automatically connect the client to IoTHub when a method is
            invoked which requires a connection to be established. (Default: True)
        :param bool connection_retry: Attempt to re-establish a dropped connection (Default: True)
        :param int connection_retry_interval: Interval, in seconds, between attempts to
            re-establish a dropped connection (Default: 10)
        :param bool ensure_desired_properties: Ensure the most recent desired properties patch has
            been received upon re-connections (Default:True)

        :raises: TypeError if given an unsupported parameter.
        :raises: ValueError if the sastoken parameter is invalid.
        """
        # Ensure no invalid kwargs were passed by the user
        excluded_kwargs = ["sastoken_ttl"]
        _validate_kwargs(exclude=excluded_kwargs, **kwargs)

        # Create SasToken object from string
        try:
            sastoken_o = st.NonRenewableSasToken(sastoken)
        except st.SasTokenError as e:
            new_err = ValueError("Invalid SasToken provided")
            new_err.__cause__ = e
            raise new_err
        # Extract values from SasToken
        vals = _extract_sas_uri_values(sastoken_o.resource_uri)
        if cls.__name__ == "IoTHubDeviceClient" and vals["module_id"]:
            raise ValueError("Provided SasToken is for a module")
        if cls.__name__ == "IoTHubModuleClient" and not vals["module_id"]:
            raise ValueError("Provided SasToken is for a device")
        if sastoken_o.expiry_time < int(time.time()):
            raise ValueError("Provided SasToken has already expired")
        # Pipeline Config setup
        config_kwargs = _get_config_kwargs(**kwargs)
        pipeline_configuration = pipeline.IoTHubPipelineConfig(
            device_id=vals["device_id"],
            module_id=vals["module_id"],
            hostname=vals["hostname"],
            sastoken=sastoken_o,
            **config_kwargs,
        )
        if cls.__name__ == "IoTHubDeviceClient":
            pipeline_configuration.blob_upload = True  # Blob Upload is a feature on Device Clients

        # Pipeline setup
        http_pipeline = pipeline.HTTPPipeline(pipeline_configuration)
        mqtt_pipeline = pipeline.MQTTPipeline(pipeline_configuration)

        return cls(mqtt_pipeline, http_pipeline)

    @abc.abstractmethod
    def shutdown(self) -> None:
        pass

    @abc.abstractmethod
    def connect(self) -> None:
        pass

    @abc.abstractmethod
    def disconnect(self) -> None:
        pass

    @abc.abstractmethod
    def update_sastoken(self, sastoken: str) -> None:
        pass

    @abc.abstractmethod
    def send_message(self, message: Union[Message, str]) -> None:
        pass

    @abc.abstractmethod
    def receive_method_request(self, method_name: Optional[str] = None) -> None:
        pass

    @abc.abstractmethod
    def send_method_response(self, method_response: MethodResponse) -> None:
        pass

    @abc.abstractmethod
    def get_twin(self) -> Twin:
        pass

    @abc.abstractmethod
    def patch_twin_reported_properties(self, reported_properties_patch: TwinPatch) -> None:
        pass

    @abc.abstractmethod
    def receive_twin_desired_properties_patch(self) -> TwinPatch:
        pass

    @property
    def connected(self) -> bool:
        """
        Read-only property to indicate if the transport is connected or not.
        """
        return self._mqtt_pipeline.connected

    @property
    def on_connection_state_change(self) -> FunctionOrCoroutine[[None], None]:
        """The handler function or coroutine that will be called when the connection state changes.

        The function or coroutine definition should take no positional arguments.
        """
        if self._handler_manager is not None:
            return self._handler_manager.on_connection_state_change

    @on_connection_state_change.setter
    def on_connection_state_change(self, value: FunctionOrCoroutine[[None], None]) -> None:
        if self._handler_manager is not None:
            self._handler_manager.on_connection_state_change = value

    @property
    def on_new_sastoken_required(self) -> FunctionOrCoroutine[[None], None]:
        """The handler function or coroutine that will be called when the client requires a new
        SAS token. This will happen approximately 2 minutes before the SAS Token expires.
        On Windows platforms, if the lifespan exceeds approximately 49 days, a new token will
        be required after those 49 days regardless of how long the SAS lifespan is.

        Note that this handler is ONLY necessary when using a client created via the
        .create_from_sastoken() method.

        The new token can be provided in your function or coroutine via use of the client's
        .update_sastoken() method.

        The function or coroutine definition should take no positional arguments.
        """
        if self._handler_manager is not None:
            return self._handler_manager.on_new_sastoken_required

    @on_new_sastoken_required.setter
    def on_new_sastoken_required(self, value: FunctionOrCoroutine[[None], None]) -> None:
        if self._handler_manager is not None:
            self._handler_manager.on_new_sastoken_required = value

    @property
    def on_background_exception(self) -> FunctionOrCoroutine[[Exception], None]:
        """The handler function or coroutine will be called when a background exception occurs.

        The function or coroutine definition should take one positional argument (the exception
        object)"""
        if self._handler_manager is not None:
            return self._handler_manager.on_background_exception

    @on_background_exception.setter
    def on_background_exception(self, value: FunctionOrCoroutine[[Exception], None]) -> None:
        if self._handler_manager is not None:
            self._handler_manager.on_background_exception = value

    @abc.abstractproperty
    def on_message_received(self) -> FunctionOrCoroutine[[Message], None]:
        # Defined below on AbstractIoTHubDeviceClient / AbstractIoTHubModuleClient
        pass

    @property
    def on_method_request_received(self) -> FunctionOrCoroutine[[MethodRequest], None]:
        """The handler function or coroutine that will be called when a method request is received.

        Remember to acknowledge the method request in your function or coroutine via use of the
        client's .send_method_response() method.

        The function or coroutine definition should take one positional argument (the
        :class:`azure.iot.device.MethodRequest` object)"""
        if self._handler_manager is not None:
            return self._handler_manager.on_method_request_received

    @on_method_request_received.setter
    def on_method_request_received(self, value: FunctionOrCoroutine[[MethodRequest], None]) -> None:
        self._generic_receive_handler_setter(
            "on_method_request_received", pipeline_constant.METHODS, value
        )

    @property
    def on_twin_desired_properties_patch_received(self) -> FunctionOrCoroutine[[TwinPatch], None]:
        """The handler function or coroutine that will be called when a twin desired properties
        patch is received.

        The function or coroutine definition should take one positional argument (the twin patch
        in the form of a JSON dictionary object)"""
        if self._handler_manager is not None:
            return self._handler_manager.on_twin_desired_properties_patch_received

    @on_twin_desired_properties_patch_received.setter
    def on_twin_desired_properties_patch_received(
        self, value: FunctionOrCoroutine[[TwinPatch], None]
    ):
        self._generic_receive_handler_setter(
            "on_twin_desired_properties_patch_received", pipeline_constant.TWIN_PATCHES, value
        )


class AbstractIoTHubDeviceClient(AbstractIoTHubClient):
    @classmethod
    def create_from_x509_certificate(
        cls, x509: X509, hostname: str, device_id: str, **kwargs
    ) -> Self:
        """
        Instantiate a client using X509 certificate authentication.

        :param str hostname: Host running the IotHub.
            Can be found in the Azure portal in the Overview tab as the string hostname.
        :param x509: The complete x509 certificate object.
            To use the certificate the enrollment object needs to contain cert
            (either the root certificate or one of the intermediate CA certificates).
            If the cert comes from a CER file, it needs to be base64 encoded.
        :type x509: :class:`azure.iot.device.X509`
        :param str device_id: The ID used to uniquely identify a device in the IoTHub

        :param str server_verification_cert: Configuration Option. The trusted certificate chain.
            Necessary when using connecting to an endpoint which has a non-standard root of trust,
            such as a protocol gateway.
        :param str gateway_hostname: Configuration Option. The gateway hostname for the gateway
            device.
        :param bool websockets: Configuration Option. Default is False. Set to true if using MQTT
            over websockets.
        :param cipher: Configuration Option. Cipher suite(s) for TLS/SSL, as a string in
            "OpenSSL cipher list format" or as a list of cipher suite strings.
        :type cipher: str or list(str)
        :param str product_info: Configuration Option. Default is empty string. The string contains
            arbitrary product info which is appended to the user agent string.
        :param proxy_options: Options for sending traffic through proxy servers.
        :type proxy_options: :class:`azure.iot.device.ProxyOptions`
        :param int keep_alive: Maximum period in seconds between communications with the
            broker. If no other messages are being exchanged, this controls the
            rate at which the client will send ping messages to the broker.
            If not provided default value of 60 secs will be used.
        :param bool auto_connect: Automatically connect the client to IoTHub when a method is
            invoked which requires a connection to be established. (Default: True)
        :param bool connection_retry: Attempt to re-establish a dropped connection (Default: True)
        :param int connection_retry_interval: Interval, in seconds, between attempts to
            re-establish a dropped connection (Default: 10)
        :param bool ensure_desired_properties: Ensure the most recent desired properties patch has
            been received upon re-connections (Default:True)

        :raises: TypeError if given an unsupported parameter.

        :returns: An instance of an IoTHub client that uses an X509 certificate for authentication.
        """
        # Ensure no invalid kwargs were passed by the user
        excluded_kwargs = ["sastoken_ttl"]
        _validate_kwargs(exclude=excluded_kwargs, **kwargs)

        # Pipeline Config setup
        config_kwargs = _get_config_kwargs(**kwargs)
        pipeline_configuration = pipeline.IoTHubPipelineConfig(
            device_id=device_id, hostname=hostname, x509=x509, **config_kwargs
        )
        pipeline_configuration.blob_upload = True  # Blob Upload is a feature on Device Clients

        # Pipeline setup
        http_pipeline = pipeline.HTTPPipeline(pipeline_configuration)
        mqtt_pipeline = pipeline.MQTTPipeline(pipeline_configuration)

        return cls(mqtt_pipeline, http_pipeline)

    @classmethod
    def create_from_symmetric_key(
        cls, symmetric_key: str, hostname: str, device_id: str, **kwargs
    ) -> Self:
        """
        Instantiate a client using symmetric key authentication.

        :param symmetric_key: The symmetric key.
        :param str hostname: Host running the IotHub.
            Can be found in the Azure portal in the Overview tab as the string hostname.
        :param device_id: The device ID

        :param str server_verification_cert: Configuration Option. The trusted certificate chain.
            Necessary when using connecting to an endpoint which has a non-standard root of trust,
            such as a protocol gateway.
        :param str gateway_hostname: Configuration Option. The gateway hostname for the gateway
            device.
        :param bool websockets: Configuration Option. Default is False. Set to true if using MQTT
            over websockets.
        :param cipher: Configuration Option. Cipher suite(s) for TLS/SSL, as a string in
            "OpenSSL cipher list format" or as a list of cipher suite strings.
        :type cipher: str or list(str)
        :param str product_info: Configuration Option. Default is empty string. The string contains
            arbitrary product info which is appended to the user agent string.
        :param proxy_options: Options for sending traffic through proxy servers.
        :type proxy_options: :class:`azure.iot.device.ProxyOptions`
        :param int sastoken_ttl: The time to live (in seconds) for the created SasToken used for
            authentication. Default is 3600 seconds (1 hour)
        :param int keep_alive: Maximum period in seconds between communications with the
            broker. If no other messages are being exchanged, this controls the
            rate at which the client will send ping messages to the broker.
            If not provided default value of 60 secs will be used.
        :param bool auto_connect: Automatically connect the client to IoTHub when a method is
            invoked which requires a connection to be established. (Default: True)
        :param bool connection_retry: Attempt to re-establish a dropped connection (Default: True)
        :param int connection_retry_interval: Interval, in seconds, between attempts to
            re-establish a dropped connection (Default: 10)
        :param bool ensure_desired_properties: Ensure the most recent desired properties patch has
            been received upon re-connections (Default:True)

        :raises: TypeError if given an unsupported parameter.
        :raises: ValueError if the provided parameters are invalid.

        :return: An instance of an IoTHub client that uses a symmetric key for authentication.
        """
        # Ensure no invalid kwargs were passed by the user
        _validate_kwargs(**kwargs)

        # Create SasToken
        uri = _form_sas_uri(hostname=hostname, device_id=device_id)
        signing_mechanism = auth.SymmetricKeySigningMechanism(key=symmetric_key)
        token_ttl = kwargs.get("sastoken_ttl", 3600)
        try:
            sastoken = st.RenewableSasToken(uri, signing_mechanism, ttl=token_ttl)
        except st.SasTokenError as e:
            new_err = ValueError("Could not create a SasToken using provided values")
            new_err.__cause__ = e
            raise new_err

        # Pipeline Config setup
        config_kwargs = _get_config_kwargs(**kwargs)
        pipeline_configuration = pipeline.IoTHubPipelineConfig(
            device_id=device_id, hostname=hostname, sastoken=sastoken, **config_kwargs
        )
        pipeline_configuration.blob_upload = True  # Blob Upload is a feature on Device Clients

        # Pipeline setup
        http_pipeline = pipeline.HTTPPipeline(pipeline_configuration)
        mqtt_pipeline = pipeline.MQTTPipeline(pipeline_configuration)

        return cls(mqtt_pipeline, http_pipeline)

    @abc.abstractmethod
    def receive_message(self) -> Message:
        pass

    @abc.abstractmethod
    def get_storage_info_for_blob(self, blob_name: str) -> Dict[str, Any]:
        pass

    @abc.abstractmethod
    def notify_blob_upload_status(
        self, correlation_id: str, is_success: bool, status_code: int, status_description: str
    ) -> None:
        pass

    @property
    def on_message_received(self) -> FunctionOrCoroutine[[Message], None]:
        """The handler function or coroutine that will be called when a message is received.

        The function or coroutine definition should take one positional argument (the
        :class:`azure.iot.device.Message` object)"""
        if self._handler_manager is not None:
            return self._handler_manager.on_message_received

    @on_message_received.setter
    def on_message_received(self, value: FunctionOrCoroutine[[Message], None]):
        self._generic_receive_handler_setter(
            "on_message_received", pipeline_constant.C2D_MSG, value
        )


class AbstractIoTHubModuleClient(AbstractIoTHubClient):
    @classmethod
    def create_from_edge_environment(cls, **kwargs) -> Self:
        """
        Instantiate the client from the IoT Edge environment.

        This method can only be run from inside an IoT Edge container, or in a debugging
        environment configured for Edge development (e.g. Visual Studio, Visual Studio Code)

        :param bool websockets: Configuration Option. Default is False. Set to true if using MQTT
            over websockets.
        :param cipher: Configuration Option. Cipher suite(s) for TLS/SSL, as a string in
            "OpenSSL cipher list format" or as a list of cipher suite strings.
        :type cipher: str or list(str)
        :param str product_info: Configuration Option. Default is empty string. The string contains
            arbitrary product info which is appended to the user agent string.
        :param proxy_options: Options for sending traffic through proxy servers.
        :type proxy_options: :class:`azure.iot.device.ProxyOptions`
        :param int sastoken_ttl: The time to live (in seconds) for the created SasToken used for
            authentication. Default is 3600 seconds (1 hour)
        :param int keep_alive: Maximum period in seconds between communications with the
            broker. If no other messages are being exchanged, this controls the
            rate at which the client will send ping messages to the broker.
            If not provided default value of 60 secs will be used.
        :param bool auto_connect: Automatically connect the client to IoTHub when a method is
            invoked which requires a connection to be established. (Default: True)
        :param bool connection_retry: Attempt to re-establish a dropped connection (Default: True)
        :param int connection_retry_interval: Interval, in seconds, between attempts to
            re-establish a dropped connection (Default: 10)

        :raises: OSError if the IoT Edge container is not configured correctly.
        :raises: ValueError if debug variables are invalid.
        :raises: TypeError if given an unsupported parameter.

        :returns: An instance of an IoTHub client that uses the IoT Edge environment for
            authentication.
        """
        # Ensure no invalid kwargs were passed by the user
        excluded_kwargs = ["server_verification_cert", "gateway_hostname"]
        _validate_kwargs(exclude=excluded_kwargs, **kwargs)

        # First try the regular Edge container variables
        try:
            hostname = os.environ["IOTEDGE_IOTHUBHOSTNAME"]
            device_id = os.environ["IOTEDGE_DEVICEID"]
            module_id = os.environ["IOTEDGE_MODULEID"]
            gateway_hostname = os.environ["IOTEDGE_GATEWAYHOSTNAME"]
            module_generation_id = os.environ["IOTEDGE_MODULEGENERATIONID"]
            workload_uri = os.environ["IOTEDGE_WORKLOADURI"]
            api_version = os.environ["IOTEDGE_APIVERSION"]
        except KeyError:
            # As a fallback, try the Edge local dev variables for debugging.
            # These variables are set by VS/VS Code in order to allow debugging
            # of Edge application code in a non-Edge dev environment.
            try:
                connection_string = os.environ["EdgeHubConnectionString"]
                ca_cert_filepath = os.environ["EdgeModuleCACertificateFile"]
            except KeyError as e:
                new_err = OSError("IoT Edge environment not configured correctly")
                new_err.__cause__ = e
                raise new_err

            # Read the certificate file to pass it on as a string
            # TODO: variant server_verification_cert file vs data object that would remove the need for this file open
            try:
                with io.open(ca_cert_filepath, mode="r") as ca_cert_file:
                    server_verification_cert = ca_cert_file.read()
            except FileNotFoundError:
                raise
            except OSError as e:
                raise ValueError("Invalid CA certificate file") from e

            # Extract config values from connection string
            connection_string = cs.ConnectionString(connection_string)
            try:
                device_id = connection_string[cs.DEVICE_ID]
                module_id = connection_string[cs.MODULE_ID]
                hostname = connection_string[cs.HOST_NAME]
                gateway_hostname = connection_string[cs.GATEWAY_HOST_NAME]
            except KeyError:
                raise ValueError("Invalid Connection String")

            # Use Symmetric Key authentication for local dev experience.
            signing_mechanism = auth.SymmetricKeySigningMechanism(
                key=connection_string[cs.SHARED_ACCESS_KEY]
            )

        else:
            # Use an HSM for authentication in the general case
            hsm = edge_hsm.IoTEdgeHsm(
                module_id=module_id,
                generation_id=module_generation_id,
                workload_uri=workload_uri,
                api_version=api_version,
            )
            try:
                server_verification_cert = hsm.get_certificate()
            except edge_hsm.IoTEdgeError as e:
                new_err = OSError("Unexpected failure in IoTEdge")
                new_err.__cause__ = e
                raise new_err
            signing_mechanism = hsm

        # Create SasToken
        uri = _form_sas_uri(hostname=hostname, device_id=device_id, module_id=module_id)
        token_ttl = kwargs.get("sastoken_ttl", 3600)
        try:
            sastoken = st.RenewableSasToken(uri, signing_mechanism, ttl=token_ttl)
        except st.SasTokenError as e:
            new_val_err = ValueError(
                "Could not create a SasToken using the values provided, or in the Edge environment"
            )
            new_val_err.__cause__ = e
            raise new_val_err

        # Pipeline Config setup
        config_kwargs = _get_config_kwargs(**kwargs)
        pipeline_configuration = pipeline.IoTHubPipelineConfig(
            device_id=device_id,
            module_id=module_id,
            hostname=hostname,
            gateway_hostname=gateway_hostname,
            sastoken=sastoken,
            server_verification_cert=server_verification_cert,
            **config_kwargs,
        )
        pipeline_configuration.method_invoke = (
            True  # Method Invoke is allowed on modules created from edge environment
        )

        # Pipeline setup
        http_pipeline = pipeline.HTTPPipeline(pipeline_configuration)
        mqtt_pipeline = pipeline.MQTTPipeline(pipeline_configuration)

        return cls(mqtt_pipeline, http_pipeline)

    @classmethod
    def create_from_x509_certificate(
        cls, x509: X509, hostname: str, device_id: str, module_id: str, **kwargs
    ) -> Self:
        """
        Instantiate a client using X509 certificate authentication.

        :param str hostname: Host running the IotHub.
            Can be found in the Azure portal in the Overview tab as the string hostname.
        :param x509: The complete x509 certificate object.
            To use the certificate the enrollment object needs to contain cert
            (either the root certificate or one of the intermediate CA certificates).
            If the cert comes from a CER file, it needs to be base64 encoded.
        :type x509: :class:`azure.iot.device.X509`
        :param str device_id: The ID used to uniquely identify a device in the IoTHub
        :param str module_id: The ID used to uniquely identify a module on a device on the IoTHub.

        :param str server_verification_cert: Configuration Option. The trusted certificate chain.
            Necessary when using connecting to an endpoint which has a non-standard root of trust,
            such as a protocol gateway.
        :param str gateway_hostname: Configuration Option. The gateway hostname for the gateway
            device.
        :param bool websockets: Configuration Option. Default is False. Set to true if using MQTT
            over websockets.
        :param cipher: Configuration Option. Cipher suite(s) for TLS/SSL, as a string in
            "OpenSSL cipher list format" or as a list of cipher suite strings.
        :type cipher: str or list(str)
        :param str product_info: Configuration Option. Default is empty string. The string contains
            arbitrary product info which is appended to the user agent string.
        :param proxy_options: Options for sending traffic through proxy servers.
        :type proxy_options: :class:`azure.iot.device.ProxyOptions`
        :param int keep_alive: Maximum period in seconds between communications with the
            broker. If no other messages are being exchanged, this controls the
            rate at which the client will send ping messages to the broker.
            If not provided default value of 60 secs will be used.
        :param bool auto_connect: Automatically connect the client to IoTHub when a method is
            invoked which requires a connection to be established. (Default: True)
        :param bool connection_retry: Attempt to re-establish a dropped connection (Default: True)
        :param int connection_retry_interval: Interval, in seconds, between attempts to
            re-establish a dropped connection (Default: 10)
        :param bool ensure_desired_properties: Ensure the most recent desired properties patch has
            been received upon re-connections (Default:True)

        :raises: TypeError if given an unsupported parameter.

        :returns: An instance of an IoTHub client that uses an X509 certificate for authentication.
        """
        # Ensure no invalid kwargs were passed by the user
        excluded_kwargs = ["sastoken_ttl"]
        _validate_kwargs(exclude=excluded_kwargs, **kwargs)

        # Pipeline Config setup
        config_kwargs = _get_config_kwargs(**kwargs)
        pipeline_configuration = pipeline.IoTHubPipelineConfig(
            device_id=device_id, module_id=module_id, hostname=hostname, x509=x509, **config_kwargs
        )

        # Pipeline setup
        http_pipeline = pipeline.HTTPPipeline(pipeline_configuration)
        mqtt_pipeline = pipeline.MQTTPipeline(pipeline_configuration)
        return cls(mqtt_pipeline, http_pipeline)

    @abc.abstractmethod
    def send_message_to_output(self, message: Union[Message, str], output_name: str) -> None:
        pass

    @abc.abstractmethod
    def receive_message_on_input(self, input_name: str) -> Message:
        pass

    @abc.abstractmethod
    def invoke_method(
        self, method_params: dict, device_id: str, module_id: Optional[str] = None
    ) -> None:
        pass

    @property
    def on_message_received(self) -> FunctionOrCoroutine[[Message], Any]:
        """The handler function or coroutine that will be called when an input message is received.

        The function definition or coroutine should take one positional argument (the
        :class:`azure.iot.device.Message` object)"""
        if self._handler_manager is not None:
            return self._handler_manager.on_message_received

    @on_message_received.setter
    def on_message_received(self, value: FunctionOrCoroutine[[Message], Any]) -> None:
        self._generic_receive_handler_setter(
            "on_message_received", pipeline_constant.INPUT_MSG, value
        )



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


class PipelineEvent(object):
    """
    A base class for data objects representing events that travels up the pipeline.

    PipelineEvent objects are used for anything that happens inside the pipeline that
    cannot be attributed to a specific operation, such as a spontaneous disconnect.

    PipelineEvents flow up the pipeline until they reach the client.  Every stage
    has the opportunity to handle a given event.  If they don't handle it, they
    should pass it up to the next stage (this is the default behavior).  Stages
    have the opportunity to tie a PipelineEvent to a PipelineOperation object
    if they are waiting for a response for that particular operation.

    :ivar name: The name of the event.  This is used primarily for logging
    :type name: str
    """

    def __init__(self):
        """
        Initializer for PipelineEvent objects.
        """
        if self.__class__ == PipelineEvent:
            raise TypeError(
                "Cannot instantiate PipelineEvent object.  You need to use a derived class"
            )
        self.name = self.__class__.__name__


class ResponseEvent(PipelineEvent):
    """
    A PipelineEvent object which is the second part of an RequestAndResponseOperation operation
    (the response).  The RequestAndResponseOperation represents the common operation of sending
    a request to iothub with a request_id ($rid) value and waiting for a response with
    the same $rid value.  This convention is used by both Twin and Provisioning features.

    The response represented by this event has not yet been matched to the corresponding
    RequestOperation operation.  That matching is done by the CoordinateRequestAndResponseStage
    stage which takes the contents of this event and puts it into the RequestAndResponseOperation
    operation with the matching $rid value.

    :ivar request_id: The request ID which will eventually be used to match a RequestOperation
      operation to this event.
    :type request_id: str
    :ivar status_code: The status code returned by the response.  Any value under 300 is
        considered success.
    :type status_code: int
    :ivar response_body: The body of the response.
    :type response_body: str
    :ivar retry_after: A retry interval value that was extracted from the topic.
    :type retry_after: int
    """

    def __init__(self, request_id, status_code, response_body, retry_after=None):
        super().__init__()
        self.request_id = request_id
        self.status_code = status_code
        self.response_body = response_body
        self.retry_after = retry_after


class ConnectedEvent(PipelineEvent):
    """
    A PipelineEvent object indicating a connection has been established.
    """

    pass


class DisconnectedEvent(PipelineEvent):
    """
    A PipelineEvent object indicating a connection has been dropped.
    """

    pass


class NewSasTokenRequiredEvent(PipelineEvent):
    """
    A PipelineEvent object indicating that a new SasToken must be provided.
    """

    pass


class BackgroundExceptionEvent(PipelineEvent):
    """
    An exception was raised in a background thread
    """

    def __init__(self, e):
        super().__init__()
        self.e = e



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
from datetime import date
import urllib

logger = logging.getLogger(__name__)

# NOTE: Whenever using standard URL encoding via the urllib.parse.quote() API
# make sure to specify that there are NO safe values (e.g. safe=""). By default
# "/" is skipped in encoding, and that is not desirable.
#
# DO NOT use urllib.parse.quote_plus(), as it turns ' ' characters into '+',
# which is invalid for MQTT publishes.
#
# DO NOT use urllib.parse.unquote_plus(), as it turns '+' characters into ' ',
# which is also invalid.


# NOTE (Oct 2020): URL encoding policy is currently inconsistent in this module due to restrictions
# with the Hub, as Hub does not do URL decoding on most values.
# (see: https://github.com/Azure/azure-iot-sdk-python/wiki/URL-Encoding-(MQTT)).
# Currently, as much as possible is URL encoded while keeping in line with the policy outlined
# in the above linked wiki article. This is to say that Device ID and Module ID are never
# encoded, however other values are. By convention, it's probably fine to be encoding/decoding most
# values that are not Device ID or Module ID, since it won't make a difference in production as
# the narrow range of acceptable values for, say, status code, or request ID don't contain any
# characters that require URL encoding/decoding in the first place. Thus it doesn't break on Hub,
# but it's still done here as a client-side best practice - Hub will eventually be doing a new API
# that does correctly URL encode/decode all values, so it's not good to roll back more than
# is currently necessary to avoid errors.


def _get_topic_base(device_id, module_id=None):
    """
    return the string that is at the beginning of all topics for this
    device/module
    """

    # NOTE: Neither Device ID nor Module ID should be URL encoded in a topic string.
    # See the repo wiki article for details:
    # https://github.com/Azure/azure-iot-sdk-python/wiki/URL-Encoding-(MQTT)
    topic = "devices/" + str(device_id)
    if module_id:
        topic = topic + "/modules/" + str(module_id)
    return topic


def get_c2d_topic_for_subscribe(device_id):
    """
    :return: The topic for cloud to device messages.It is of the format
    "devices/<deviceid>/messages/devicebound/#"
    """
    return _get_topic_base(device_id) + "/messages/devicebound/#"


def get_input_topic_for_subscribe(device_id, module_id):
    """
    :return: The topic for input messages. It is of the format
    "devices/<deviceId>/modules/<moduleId>/inputs/#"
    """
    return _get_topic_base(device_id, module_id) + "/inputs/#"


def get_method_topic_for_subscribe():
    """
    :return: The topic for ALL incoming methods. It is of the format
    "$iothub/methods/POST/#"
    """
    return "$iothub/methods/POST/#"


def get_twin_response_topic_for_subscribe():
    """
    :return: The topic for ALL incoming twin responses. It is of the format
    "$iothub/twin/res/#"
    """
    return "$iothub/twin/res/#"


def get_twin_patch_topic_for_subscribe():
    """
    :return: The topic for ALL incoming twin patches. It is of the format
    "$iothub/twin/PATCH/properties/desired/#
    """
    return "$iothub/twin/PATCH/properties/desired/#"


def get_telemetry_topic_for_publish(device_id, module_id):
    """
    return the topic string used to publish telemetry
    """
    return _get_topic_base(device_id, module_id) + "/messages/events/"


def get_method_topic_for_publish(request_id, status):
    """
    :return: The topic for publishing method responses. It is of the format
    "$iothub/methods/res/<status>/?$rid=<requestId>
    """
    return "$iothub/methods/res/{status}/?$rid={request_id}".format(
        status=urllib.parse.quote(str(status), safe=""),
        request_id=urllib.parse.quote(str(request_id), safe=""),
    )


# NOTE: Consider splitting this into separate logic for Twin Requests / Twin Patches
# This is the only method that is shared. Would probably simplify code if it was split.
# Please consider refactoring.
def get_twin_topic_for_publish(method, resource_location, request_id):
    """
    :return: The topic for publishing twin requests / patches. It is of the format
    "$iothub/twin/<method><resourceLocation>?$rid=<requestId>
    """
    return "$iothub/twin/{method}{resource_location}?$rid={request_id}".format(
        method=method,
        resource_location=resource_location,
        request_id=urllib.parse.quote(str(request_id), safe=""),
    )


def is_c2d_topic(topic, device_id):
    """
    Topics for c2d message are of the following format:
    devices/<deviceId>/messages/devicebound
    :param topic: The topic string
    """
    # Device ID is not URL encoded in a topic string
    # See the repo wiki article for details:
    # https://github.com/Azure/azure-iot-sdk-python/wiki/URL-Encoding-(MQTT)
    if "devices/{}/messages/devicebound".format(device_id) in topic:
        return True
    return False


def is_input_topic(topic, device_id, module_id):
    """
    Topics for inputs are of the following format:
    devices/<deviceId>/modules/<moduleId>/inputs/<inputName>
    :param topic: The topic string
    """
    if not device_id or not module_id:
        return False
    # NOTE: Neither Device ID nor Module ID are URL encoded in a topic string.
    # See the repo wiki article for details:
    # https://github.com/Azure/azure-iot-sdk-python/wiki/URL-Encoding-(MQTT)
    if "devices/{}/modules/{}/inputs/".format(device_id, module_id) in topic:
        return True
    return False


def is_method_topic(topic):
    """
    Topics for methods are of the following format:
    "$iothub/methods/POST/{method name}/?$rid={request id}"

    :param str topic: The topic string.
    """
    if "$iothub/methods/POST" in topic:
        return True
    return False


def is_twin_response_topic(topic):
    """Topics for twin responses are of the following format:
    $iothub/twin/res/{status}/?$rid={rid}

    :param str topic: The topic string
    """
    return topic.startswith("$iothub/twin/res/")


def is_twin_desired_property_patch_topic(topic):
    """Topics for twin desired property patches are of the following format:
    $iothub/twin/PATCH/properties/desired

    :param str topic: The topic string
    """
    return topic.startswith("$iothub/twin/PATCH/properties/desired")


def get_input_name_from_topic(topic):
    """
    Extract the input channel from the topic name
    Topics for inputs are of the following format:
    devices/<deviceId>/modules/<moduleId>/inputs/<inputName>

    :param topic: The topic string
    """
    parts = topic.split("/")
    if len(parts) > 5 and parts[4] == "inputs":
        return urllib.parse.unquote(parts[5])
    else:
        raise ValueError("topic has incorrect format")


def get_method_name_from_topic(topic):
    """
    Extract the method name from the method topic.
    Topics for methods are of the following format:
    "$iothub/methods/POST/{method name}/?$rid={request id}"

    :param str topic: The topic string
    """
    parts = topic.split("/")
    if is_method_topic(topic) and len(parts) >= 4:
        return urllib.parse.unquote(parts[3])
    else:
        raise ValueError("topic has incorrect format")


def get_method_request_id_from_topic(topic):
    """
    Extract the Request ID (RID) from the method topic.
    Topics for methods are of the following format:
    "$iothub/methods/POST/{method name}/?$rid={request id}"

    :param str topic: the topic string
    :raises: ValueError if topic has incorrect format
    :returns: request id from topic string
    """
    parts = topic.split("/")
    if is_method_topic(topic) and len(parts) >= 4:

        properties = _extract_properties(topic.split("?")[1])
        return properties["rid"]
    else:
        raise ValueError("topic has incorrect format")


def get_twin_request_id_from_topic(topic):
    """
    Extract the Request ID (RID) from the twin response topic.
    Topics for twin response are in the following format:
    "$iothub/twin/res/{status}/?$rid={rid}"

    :param str topic: The topic string
    :raises: ValueError if topic has incorrect format
    :returns: request id from topic string
    """
    parts = topic.split("/")
    if is_twin_response_topic(topic) and len(parts) >= 4:
        properties = _extract_properties(topic.split("?")[1])
        return properties["rid"]
    else:
        raise ValueError("topic has incorrect format")


def get_twin_status_code_from_topic(topic):
    """
    Extract the status code from the twin response topic.
    Topics for twin response are in the following format:
    "$iothub/twin/res/{status}/?$rid={rid}"

    :param str topic: The topic string
    :raises: ValueError if the topic has incorrect format
    :returns status code from topic string
    """
    parts = topic.split("/")
    if is_twin_response_topic(topic) and len(parts) >= 4:
        return urllib.parse.unquote(parts[3])
    else:
        raise ValueError("topic has incorrect format")


def extract_message_properties_from_topic(topic, message_received):
    """
    Extract key=value pairs from custom properties and set the properties on the received message.
    For extracting values corresponding to keys the following rules are followed:-
    If there is NO "=", the value is None
    If there is "=" with no value, the value is an empty string
    For anything else the value after "=" and before `&` is considered as the proper value
    :param topic: The topic string
    :param message_received: The message received with the payload in bytes
    """

    parts = topic.split("/")
    # Input Message Topic
    if len(parts) > 4 and parts[4] == "inputs":
        if len(parts) > 6:
            properties = parts[6]
        else:
            properties = None
    # C2D Message Topic
    elif len(parts) > 3 and parts[3] == "devicebound":
        if len(parts) > 4:
            properties = parts[4]
        else:
            properties = None
    else:
        raise ValueError("topic has incorrect format")

    # We do not want to extract values corresponding to these keys
    ignored_extraction_values = ["$.to"]

    # NOTE: we cannot use urllib.parse.parse_qs because it always decodes '+' as ' ',
    # and the behavior cannot be overridden. Must parse key/value pairs manually.

    if properties:
        key_value_pairs = properties.split("&")

        for entry in key_value_pairs:
            pair = entry.split("=")
            key = urllib.parse.unquote(pair[0])
            if len(pair) > 1:
                value = urllib.parse.unquote(pair[1])
            else:  # Don't skip the key
                value = None

            if key in ignored_extraction_values:
                continue
            elif key == "$.mid":
                message_received.message_id = value
            elif key == "$.cid":
                message_received.correlation_id = value
            elif key == "$.uid":
                message_received.user_id = value
            elif key == "$.ct":
                message_received.content_type = value
            elif key == "$.ce":
                message_received.content_encoding = value
            elif key == "$.exp":
                message_received.expiry_time_utc = value
            elif key == "iothub-ack":
                message_received.ack = value
            else:
                message_received.custom_properties[key] = value


def encode_message_properties_in_topic(message_to_send, topic):
    """
    uri-encode the system properties of a message as key-value pairs on the topic with defined keys.
    Additionally if the message has user defined properties, the property keys and values shall be
    uri-encoded and appended at the end of the above topic with the following convention:
    '<key>=<value>&<key2>=<value2>&<key3>=<value3>(...)'
    :param message_to_send: The message to send
    :param topic: The topic which has not been encoded yet. For a device it looks like
    "devices/<deviceId>/messages/events/" and for a module it looks like
    "devices/<deviceId>/modules/<moduleId>/messages/events/
    :return: The topic which has been uri-encoded
    """
    system_properties = []
    if message_to_send.output_name:
        system_properties.append(("$.on", str(message_to_send.output_name)))
    if message_to_send.message_id:
        system_properties.append(("$.mid", str(message_to_send.message_id)))

    if message_to_send.correlation_id:
        system_properties.append(("$.cid", str(message_to_send.correlation_id)))

    if message_to_send.user_id:
        system_properties.append(("$.uid", str(message_to_send.user_id)))

    if message_to_send.content_type:
        system_properties.append(("$.ct", str(message_to_send.content_type)))

    if message_to_send.content_encoding:
        system_properties.append(("$.ce", str(message_to_send.content_encoding)))

    if message_to_send.iothub_interface_id:
        system_properties.append(("$.ifid", str(message_to_send.iothub_interface_id)))

    if message_to_send.expiry_time_utc:
        system_properties.append(
            (
                "$.exp",
                message_to_send.expiry_time_utc.isoformat()  # returns string
                if isinstance(message_to_send.expiry_time_utc, date)
                else message_to_send.expiry_time_utc,
            )
        )

    system_properties_encoded = urllib.parse.urlencode(
        system_properties, quote_via=urllib.parse.quote
    )
    topic += system_properties_encoded

    if message_to_send.custom_properties and len(message_to_send.custom_properties) > 0:
        if system_properties and len(system_properties) > 0:
            topic += "&"

        # Convert the custom properties to a sorted list in order to ensure the
        # resulting ordering in the topic string is consistent across versions of Python.
        # Convert to the properties to strings for safety.
        custom_prop_seq = [
            (str(i[0]), str(i[1])) for i in list(message_to_send.custom_properties.items())
        ]
        custom_prop_seq.sort()

        # Validate that string conversion has not created duplicate keys
        keys = [i[0] for i in custom_prop_seq]
        if len(keys) != len(set(keys)):
            raise ValueError("Duplicate keys in custom properties!")

        user_properties_encoded = urllib.parse.urlencode(
            custom_prop_seq, quote_via=urllib.parse.quote
        )
        topic += user_properties_encoded

    return topic


def _extract_properties(properties_str):
    """Return a dictionary of properties from a string in the format
    ${key1}={value1}&${key2}={value2}...&${keyn}={valuen}
    """
    d = {}
    kv_pairs = properties_str.split("&")

    for entry in kv_pairs:
        pair = entry.split("=")
        key = urllib.parse.unquote(pair[0]).lstrip("$")
        value = urllib.parse.unquote(pair[1])
        d[key] = value

    return d



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from azure.iot.device.common.pipeline.pipeline_ops_base import PipelineOperation


class RegisterOperation(PipelineOperation):
    """
    A PipelineOperation object which contains arguments used to send a registration request
    to an Device Provisioning Service.

    This operation is in the group of DPS operations because it is very specific to the DPS client.
    """

    def __init__(self, request_payload, registration_id, callback, registration_result=None):
        """
        Initializer for RegisterOperation objects.

        :param request_payload: The request that we are sending to the service
        :param registration_id: The registration ID is used to uniquely identify a device in the Device Provisioning Service.
        :param Function callback: The function that gets called when this operation is complete or has failed.
         The callback function must accept A PipelineOperation object which indicates the specific operation which
         has completed or failed.
        """
        super().__init__(callback=callback)
        self.request_payload = request_payload
        self.registration_id = registration_id
        self.registration_result = registration_result
        self.retry_after_timer = None
        self.polling_timer = None
        self.provisioning_timeout_timer = None


class PollStatusOperation(PipelineOperation):
    """
    A PipelineOperation object which contains arguments used to send a registration request
    to an Device Provisioning Service.

    This operation is in the group of DPS operations because it is very specific to the DPS client.
    """

    def __init__(self, operation_id, request_payload, callback, registration_result=None):
        """
        Initializer for PollStatusOperation objects.

        :param operation_id: The id of the existing operation for which the polling was started.
        :param request_payload: The request that we are sending to the service
        :param Function callback: The function that gets called when this operation is complete or has failed.
         The callback function must accept A PipelineOperation object which indicates the specific operation which
         has completed or failed.
        """
        super().__init__(callback=callback)
        self.operation_id = operation_id
        self.request_payload = request_payload
        self.registration_result = registration_result
        self.retry_after_timer = None
        self.polling_timer = None
        self.provisioning_timeout_timer = None



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from . import PipelineOperation


class MQTTPublishOperation(PipelineOperation):
    """
    A PipelineOperation object which contains arguments used to publish a specific payload on a specific topic using the MQTT protocol.

    This operation is in the group of MQTT operations because its attributes are very specific to the MQTT protocol.
    """

    def __init__(self, topic, payload, callback):
        """
        Initializer for MQTTPublishOperation objects.

        :param str topic: The name of the topic to publish to
        :param str payload: The payload to publish
        :param Function callback: The function that gets called when this operation is complete or has failed.
          The callback function must accept A PipelineOperation object which indicates the specific operation which
          has completed or failed.
        """
        super().__init__(callback=callback)
        self.topic = topic
        self.payload = payload
        self.needs_connection = True
        self.retry_timer = None


class MQTTSubscribeOperation(PipelineOperation):
    """
    A PipelineOperation object which contains arguments used to subscribe to a specific MQTT topic using the MQTT protocol.

    This operation is in the group of MQTT operations because its attributes are very specific to the MQTT protocol.
    """

    def __init__(self, topic, callback):
        """
        Initializer for MQTTSubscribeOperation objects.

        :param str topic: The name of the topic to subscribe to
        :param Function callback: The function that gets called when this operation is complete or has failed.
          The callback function must accept A PipelineOperation object which indicates the specific operation which
          has completed or failed.
        """
        super().__init__(callback=callback)
        self.topic = topic
        self.needs_connection = True
        self.timeout_timer = None
        self.retry_timer = None


class MQTTUnsubscribeOperation(PipelineOperation):
    """
    A PipelineOperation object which contains arguments used to unsubscribe from a specific MQTT topic using the MQTT protocol.

    This operation is in the group of MQTT operations because its attributes are very specific to the MQTT protocol.
    """

    def __init__(self, topic, callback):
        """
        Initializer for MQTTUnsubscribeOperation objects.

        :param str topic: The name of the topic to unsubscribe from
        :param Function callback: The function that gets called when this operation is complete or has failed.
          The callback function must accept A PipelineOperation object which indicates the specific operation which
          has completed or failed.
        """
        super().__init__(callback=callback)
        self.topic = topic
        self.needs_connection = True
        self.timeout_timer = None
        self.retry_timer = None



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from . import PipelineOperation


class HTTPRequestAndResponseOperation(PipelineOperation):
    """
    A PipelineOperation object which contains arguments used to connect to a server using the HTTP protocol.

    This operation is in the group of HTTP operations because its attributes are very specific to the HTTP protocol.
    """

    def __init__(self, method, path, headers, body, query_params, callback):
        """
        Initializer for HTTPPublishOperation objects.
        :param str method: The HTTP method used in the request
        :param str path: The path to be used in the request url
        :param dict headers: The headers to be used in the HTTP request
        :param str body: The body to be provided with the HTTP request
        :param str query_params: The query parameters to be used in the request url
        :param Function callback: The function that gets called when this operation is complete or has failed.
            The callback function must accept A PipelineOperation object which indicates the specific operation which
            has completed or failed.
        """
        super().__init__(callback=callback)
        self.method = method
        self.path = path
        self.headers = headers
        self.body = body
        self.query_params = query_params
        self.status_code = None
        self.response_body = None
        self.reason = None



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
from azure.iot.device.common.pipeline.config import BasePipelineConfig

logger = logging.getLogger(__name__)


class ProvisioningPipelineConfig(BasePipelineConfig):
    """A class for storing all configurations/options for Provisioning clients in the Azure IoT Python Device Client Library."""

    def __init__(self, hostname, registration_id, id_scope, **kwargs):
        """Initializer for ProvisioningPipelineConfig which passes all unrecognized keyword-args down to BasePipelineConfig
        to be evaluated. This stacked options setting is to allow for unique configuration options to exist between the
        multiple clients, while maintaining a base configuration class with shared config options.

        :param str hostname: The hostname of the Provisioning hub instance to connect to
        :param str registration_id: The device registration identity being provisioned
        :param str id_scope: The identity of the provisioning service being used
        """
        super().__init__(hostname=hostname, **kwargs)

        # Provisioning Connection Details
        self.registration_id = registration_id
        self.id_scope = id_scope



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module defines an exception surface, exposed as part of the azure.iot.device library API"""

# Currently, we are redefining many lower level exceptions in this file, in order to present an API
# surface that will be consistent and unchanging (even though lower level exceptions may change).
# Potentially, this could be somewhat relaxed in the future as the design solidifies.

# ~~~ EXCEPTIONS ~~~


class OperationCancelled(Exception):
    """An operation was cancelled"""

    pass


class OperationTimeout(Exception):
    """An operation timed out"""

    pass


# ~~~ CLIENT ERRORS ~~~


class ClientError(Exception):
    """Generic error for a client"""

    pass


class ConnectionFailedError(ClientError):
    """Failed to establish a connection"""

    pass


class ConnectionDroppedError(ClientError):
    """Lost connection while executing operation"""

    pass


class NoConnectionError(ClientError):
    """Operation could not be completed because no connection has been established"""

    pass


class CredentialError(ClientError):
    """Could not connect client using given credentials"""

    pass


# ~~~ SERVICE ERRORS ~~~


class ServiceError(Exception):
    """Error received from an Azure IoT service"""

    pass


# NOTE: These are not (yet) in use.
# Because of this they have been commented out to prevent confusion.

# class ArgumentError(ServiceError):
#     """Service returned 400"""

#     pass


# class UnauthorizedError(ServiceError):
#     """Service returned 401"""

#     pass


# class QuotaExceededError(ServiceError):
#     """Service returned 403"""

#     pass


# class NotFoundError(ServiceError):
#     """Service returned 404"""

#     pass


# class DeviceTimeoutError(ServiceError):
#     """Service returned 408"""

#     # TODO: is this a method call error?  If so, do we retry?
#     pass


# class DeviceAlreadyExistsError(ServiceError):
#     """Service returned 409"""

#     pass


# class InvalidEtagError(ServiceError):
#     """Service returned 412"""

#     pass


# class MessageTooLargeError(ServiceError):
#     """Service returned 413"""

#     pass


# class ThrottlingError(ServiceError):
#     """Service returned 429"""

#     pass


# class InternalServiceError(ServiceError):
#     """Service returned 500"""

#     pass


# class BadDeviceResponseError(ServiceError):
#     """Service returned 502"""

#     # TODO: is this a method invoke thing?
#     pass


# class ServiceUnavailableError(ServiceError):
#     """Service returned 503"""

#     pass


# class ServiceTimeoutError(ServiceError):
#     """Service returned 504"""

#     pass


# class FailedStatusCodeError(ServiceError):
#     """Service returned unknown status code"""

#     pass


# status_code_to_error = {
#     400: ArgumentError,
#     401: UnauthorizedError,
#     403: QuotaExceededError,
#     404: NotFoundError,
#     408: DeviceTimeoutError,
#     409: DeviceAlreadyExistsError,
#     412: InvalidEtagError,
#     413: MessageTooLargeError,
#     429: ThrottlingError,
#     500: InternalServiceError,
#     502: BadDeviceResponseError,
#     503: ServiceUnavailableError,
#     504: ServiceTimeoutError,
# }


# def error_from_status_code(status_code, message=None):
#     """
#     Return an Error object from a failed status code

#     :param int status_code: Status code returned from failed operation
#     :returns: Error object
#     """
#     if status_code in status_code_to_error:
#         return status_code_to_error[status_code](message)
#     else:
#         return FailedStatusCodeError(message)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import functools
import logging
import threading
import traceback
from concurrent.futures import ThreadPoolExecutor
from azure.iot.device.common import handle_exceptions

logger = logging.getLogger(__name__)

"""
This module contains decorators that are used to marshal code into pipeline and
callback threads and to assert that code is being called in the correct thread.

The intention of these decorators is to ensure the following:

1. All pipeline functions execute in a single thread, known as the "pipeline
  thread".  The `invoke_on_pipeline_thread` and `invoke_on_pipeline_thread_nowait`
  decorators cause the decorated function to run on the pipeline thread.

2. If the pipeline thread is busy running a different function, the invoke
  decorators will wait until that function is complete before invoking another
  function on that thread.

3. There is a different thread which is used for callbacks into user code, known
  as the the "callback thread".  This is not meant for callbacks into pipeline
  code.  Those callbacks should still execute on the pipeline thread.  The
  `invoke_on_callback_thread_nowait` decorator is used to ensure that callbacks
  execute on the callback thread.

4. Decorators which cause thread switches are used only when necessary.  The
  pipeline thread is only entered in places where we know that external code is
  calling into the pipeline (such as a client API call or a callback from a
  third-party library).  Likewise, the callback thread is only entered in places
  where we know that the pipeline is calling back into client code.

5. Exceptions raised from the pipeline thread are still able to be caught by
  the function which entered the pipeline thread.

5. Calls into the pipeline thread can either block or not block.  Blocking is used
  for cases where the caller needs a return value from the pipeline or is
  expecting to handle any errors raised from the pipeline thread.  Blocking is
  not used when the code calling into the pipeline is not waiting for a response
  and is not expecting to handle any exceptions, such as protocol library
  handlers which call into the pipeline to deliver protocol messages.

6. Calls into the callback thread could theoretically block, but we currently
  only have decorators which enter the callback thread without blocking.  This
  is done to ensure that client code does not execute on the pipeline thread and
  also to ensure that the pipeline thread is not blocked while waiting for client
  code to execute.

These decorators use concurrent.futures.Future and the ThreadPoolExecutor because:

1. The thread pooling with a pool size of 1 gives us a single thread to run all
  pipeline operations and a different (single) thread to run all callbacks.  If
  the code attempts to run a second pipeline operation (or callback) while a
  different one is running, the ThreadPoolExecutor will queue the code until the
  first call is completed.

2. The concurrent.futures.Future object properly handles both Exception and
  BaseException errors, re-raising them when the Future.result method is called.
  threading.Thread.get() was not an option because it doesn't re-raise
  BaseException errors when Thread.get is called.
"""

_executors = {}


def _get_named_executor(thread_name):
    """
    Get a ThreadPoolExecutor object with the given name.  If no such executor exists,
    this function will create on with a single worker and assign it to the provided
    name.
    """
    global _executors
    if thread_name not in _executors:
        logger.debug("Creating {} executor".format(thread_name))
        _executors[thread_name] = ThreadPoolExecutor(max_workers=1)
    return _executors[thread_name]


def _invoke_on_executor_thread(func, thread_name, block=True):
    """
    Return wrapper to run the function on a given thread.  If block==False,
    the call returns immediately without waiting for the decorated function to complete.
    If block==True, the call waits for the decorated function to complete before returning.
    """

    # Mocks and other callable objects don't have a __name__ attribute.
    # Use str() if you can't use __name__
    try:
        function_name = func.__name__
    except AttributeError:
        function_name = str(func)

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if threading.current_thread().name is not thread_name:
            logger.debug("Starting {} in {} thread".format(function_name, thread_name))

            def thread_proc():
                threading.current_thread().name = thread_name
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    if not block:
                        handle_exceptions.handle_background_exception(e)
                    else:
                        raise
                except BaseException:
                    if not block:
                        # This is truly a logger.critical condition.  Most exceptions in background threads should
                        # be handled inside the thread and should result in call to handle_background_exception
                        # if this code is hit, that means something happened which wasn't handled, therefore
                        # handle_background_exception wasn't called, therefore we need to log this at the highest
                        # level.
                        logger.critical("Unhandled exception in background thread")
                        logger.critical(
                            "This may cause the background thread to abort and may result in system instability."
                        )
                        traceback.print_exc()
                    raise

            # TODO: add a timeout here and throw exception on failure
            future = _get_named_executor(thread_name).submit(thread_proc)
            if block:
                return future.result()
            else:
                return future
        else:
            logger.debug("Already in {} thread for {}".format(thread_name, function_name))
            return func(*args, **kwargs)

    return wrapper


def invoke_on_pipeline_thread(func):
    """
    Run the decorated function on the pipeline thread.
    """
    return _invoke_on_executor_thread(func=func, thread_name="pipeline")


def invoke_on_pipeline_thread_nowait(func):
    """
    Run the decorated function on the pipeline thread, but don't wait for it to complete
    """
    return _invoke_on_executor_thread(func=func, thread_name="pipeline", block=False)


def invoke_on_callback_thread_nowait(func):
    """
    Run the decorated function on the callback thread, but don't wait for it to complete
    """
    return _invoke_on_executor_thread(func=func, thread_name="callback", block=False)


def invoke_on_http_thread_nowait(func):
    """
    Run the decorated function on the callback thread, but don't wait for it to complete
    """
    # TODO: Refactor this since this is not in the pipeline thread anymore, so we need to pull this into common.
    # Also, the max workers eventually needs to be a bigger number, so that needs to be fixed to allow for more than one HTTP Request a a time.
    return _invoke_on_executor_thread(func=func, thread_name="azure_iot_http", block=False)


def _assert_executor_thread(func, thread_name):
    """
    Decorator which asserts that the given function only gets called inside the given
    thread.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):

        assert (
            threading.current_thread().name == thread_name
        ), """
            Function {function_name} is not running inside {thread_name} thread.
            It should be. You should use invoke_on_{thread_name}_thread(_nowait) to enter the
            {thread_name} thread before calling this function.  If you're hitting this from
            inside a test function, you may need to add the fake_pipeline_thread fixture to
            your test.  (generally applied on the global pytestmark in a module) """.format(
            function_name=func.__name__, thread_name=thread_name
        )

        return func(*args, **kwargs)

    return wrapper


def runs_on_pipeline_thread(func):
    """
    Decorator which marks a function as only running inside the pipeline thread.
    """
    return _assert_executor_thread(func=func, thread_name="pipeline")


def runs_on_http_thread(func):
    """
    Decorator which marks a function as only running inside the http thread.
    """
    return _assert_executor_thread(func=func, thread_name="azure_iot_http")



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
from azure.iot.device.common.evented_callback import EventedCallback
from azure.iot.device.common.pipeline import (
    pipeline_stages_base,
    pipeline_ops_base,
    pipeline_stages_mqtt,
    pipeline_exceptions,
    pipeline_nucleus,
)
from azure.iot.device.provisioning.pipeline import (
    pipeline_stages_provisioning,
    pipeline_stages_provisioning_mqtt,
)
from azure.iot.device.provisioning.pipeline import pipeline_ops_provisioning
from azure.iot.device.provisioning.pipeline import constant as provisioning_constants

logger = logging.getLogger(__name__)


class MQTTPipeline(object):
    def __init__(self, pipeline_configuration):
        """
        Constructor for instantiating a pipeline
        :param security_client: The security client which stores credentials
        """
        self.responses_enabled = {provisioning_constants.REGISTER: False}

        # Event Handlers - Will be set by Client after instantiation of pipeline
        self.on_connected = None
        self.on_disconnected = None
        self.on_background_exception = None
        self.on_message_received = None
        self._registration_id = pipeline_configuration.registration_id

        # Contains data and information shared globally within the pipeline
        self._nucleus = pipeline_nucleus.PipelineNucleus(pipeline_configuration)

        self._pipeline = (
            #
            # The root is always the root.  By definition, it's the first stage in the pipeline.
            #
            pipeline_stages_base.PipelineRootStage(self._nucleus)
            #
            # SasTokenStage comes near the root by default because it should be as close
            # to the top of the pipeline as possible, and does not need to be after anything.
            #
            .append_stage(pipeline_stages_base.SasTokenStage())
            #
            # RegistrationStage needs to come early because this is the stage that converts registration
            # or query requests into request and response objects which are used by later stages
            #
            .append_stage(pipeline_stages_provisioning.RegistrationStage())
            #
            # PollingStatusStage needs to come after RegistrationStage because RegistrationStage counts
            # on PollingStatusStage to poll until the registration is complete.
            #
            .append_stage(pipeline_stages_provisioning.PollingStatusStage())
            #
            # CoordinateRequestAndResponseStage needs to be after RegistrationStage and PollingStatusStage
            # because these 2 stages create the request ops that CoordinateRequestAndResponseStage
            # is coordinating.  It needs to be before ProvisioningMQTTTranslationStage because that stage
            # operates on ops that CoordinateRequestAndResponseStage produces
            #
            .append_stage(pipeline_stages_base.CoordinateRequestAndResponseStage())
            #
            # ProvisioningMQTTTranslationStage comes here because this is the point where we can translate
            # all operations directly into MQTT.  After this stage, only pipeline_stages_base stages
            # are allowed because ProvisioningMQTTTranslationStage removes all the provisioning-ness from the ops
            #
            .append_stage(pipeline_stages_provisioning_mqtt.ProvisioningMQTTTranslationStage())
            #
            # AutoConnectStage comes here because only MQTT ops have the need_connection flag set
            # and this is the first place in the pipeline where we can guarantee that all network
            # ops are MQTT ops.
            #
            .append_stage(pipeline_stages_base.AutoConnectStage())
            #
            # ConnectionStateStage needs to be after AutoConnectStage because the AutoConnectStage
            # can create ConnectOperations and we (may) want to queue connection related operations
            # in the ConnectionStateStage
            #
            .append_stage(pipeline_stages_base.ConnectionStateStage())
            #
            # RetryStage needs to be near the end because it's retrying low-level MQTT operations.
            #
            .append_stage(pipeline_stages_base.RetryStage())
            #
            # OpTimeoutStage needs to be after RetryStage because OpTimeoutStage returns the timeout
            # errors that RetryStage is watching for.
            #
            .append_stage(pipeline_stages_base.OpTimeoutStage())
            #
            # MQTTTransportStage needs to be at the very end of the pipeline because this is where
            # operations turn into network traffic
            #
            .append_stage(pipeline_stages_mqtt.MQTTTransportStage())
        )

        def _on_pipeline_event(event):
            # error because no events should
            logger.debug("Dropping unknown pipeline event {}".format(event.name))

        def _on_connected():
            if self.on_connected:
                self.on_connected("connected")

        def _on_disconnected():
            if self.on_disconnected:
                self.on_disconnected("disconnected")

        def _on_background_exception():
            if self.on_background_exception:
                self.on_background_exception

        self._pipeline.on_pipeline_event_handler = _on_pipeline_event
        self._pipeline.on_connected_handler = _on_connected
        self._pipeline.on_disconnected_handler = _on_disconnected

        callback = EventedCallback()
        op = pipeline_ops_base.InitializePipelineOperation(callback=callback)

        self._pipeline.run_op(op)
        callback.wait_for_completion()

        # Set the running flag
        self._running = True

    def _verify_running(self):
        if not self._running:
            raise pipeline_exceptions.PipelineNotRunning(
                "Cannot execute method - Pipeline is not running"
            )

    def shutdown(self, callback):
        """Shut down the pipeline and clean up any resources.

        Once shut down, making any further calls on the pipeline will result in a
        PipelineNotRunning exception being raised.

        There is currently no way to resume pipeline functionality once shutdown has occurred.

        :raises: :class:`azure.iot.device.provisioning.pipeline.exceptions.PipelineNotRunning` if the
            pipeline has already been shut down

        The shutdown process itself is not expected to fail under any normal condition, but if it
        does, exceptions are not "raised", but rather, returned via the "error" parameter when
        invoking "callback".
        """
        self._verify_running()
        logger.debug("Commencing shutdown of pipeline")

        def on_complete(op, error):
            if not error:
                # Only set the pipeline to not be running if the op was successful
                self._running = False
            callback(error=error)

        # NOTE: While we do run this operation, its functionality is incomplete. Some stages still
        # need a response to this operation implemented. Additionally, there are other pipeline
        # constructs other than Stages (e.g. Operations) which may have timers attached. These are
        # lesser issues, but should be addressed at some point.
        # TODO: Truly complete the shutdown implementation
        self._pipeline.run_op(pipeline_ops_base.ShutdownPipelineOperation(callback=on_complete))

    def connect(self, callback=None):
        """
        Connect to the service.

        :param callback: callback which is called when the connection to the service is complete.

        :raises: :class:`azure.iot.device.provisioning.pipeline.exceptions.PipelineNotRunning` if the
            pipeline has already been shut down

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.provisioning.pipeline.exceptions.ConnectionFailedError`
        :raises: :class:`azure.iot.device.provisioning.pipeline.exceptions.ConnectionDroppedError`
        :raises: :class:`azure.iot.device.provisioning.pipeline.exceptions.UnauthorizedError`
        :raises: :class:`azure.iot.device.provisioning.pipeline.exceptions.ProtocolClientError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.OperationTimeout`
        """
        self._verify_running()
        logger.debug("connect called")

        def pipeline_callback(op, error):
            callback(error=error)

        self._pipeline.run_op(pipeline_ops_base.ConnectOperation(callback=pipeline_callback))

    def disconnect(self, callback=None):
        """
        Disconnect from the service.

        :param callback: callback which is called when the connection to the service has been disconnected

        :raises: :class:`azure.iot.device.provisioning.pipeline.exceptions.PipelineNotRunning` if the
            pipeline has already been shut down

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`
        """
        self._verify_running()
        logger.debug("disconnect called")

        def pipeline_callback(op, error):
            callback(error=error)

        self._pipeline.run_op(pipeline_ops_base.DisconnectOperation(callback=pipeline_callback))

    # NOTE: Currently, this operation will retry itself indefinitely in the case of timeout
    def enable_responses(self, callback=None):
        """
        Enable response from the DPS service by subscribing to the appropriate topics.

        :raises: :class:`azure.iot.device.provisioning.pipeline.exceptions.PipelineNotRunning` if the
            pipeline has already been shut down

        :param callback: callback which is called when responses are enabled
        """
        self._verify_running()
        logger.debug("enable_responses called")

        self.responses_enabled[provisioning_constants.REGISTER] = True

        def pipeline_callback(op, error):
            callback(error=error)

        self._pipeline.run_op(
            pipeline_ops_base.EnableFeatureOperation(
                feature_name=provisioning_constants.REGISTER, callback=pipeline_callback
            )
        )

    def register(self, payload=None, callback=None):
        """
        Register to the device provisioning service.
        :param payload: Payload that can be sent with the registration request.
        :param callback: callback which is called when the registration is done.

        :raises: :class:`azure.iot.device.provisioning.pipeline.exceptions.PipelineNotRunning` if the
            pipeline has already been shut down

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.NoConnectionError`
        :raises: :class:`azure.iot.device.provisioning.pipeline.exceptions.ProtocolClientError

        The following exceptions can be returned via the "error" parameter only if auto-connect
        is enabled in the pipeline configuration:

        :raises: :class:`azure.iot.device.provisioning.pipeline.exceptions.ConnectionFailedError`
        :raises: :class:`azure.iot.device.provisioning.pipeline.exceptions.ConnectionDroppedError`
        :raises: :class:`azure.iot.device.provisioning.pipeline.exceptions.UnauthorizedError``
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.OperationTimeout`
        """
        self._verify_running()

        def on_complete(op, error):
            # TODO : Apparently when its failed we can get result as well as error.
            if error:
                callback(error=error, result=None)
            else:
                callback(result=op.registration_result)

        self._pipeline.run_op(
            pipeline_ops_provisioning.RegisterOperation(
                request_payload=payload, registration_id=self._registration_id, callback=on_complete
            )
        )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module contains a manager for inboxes."""

import logging

logger = logging.getLogger(__name__)


class InboxManager(object):
    """Manages the various Inboxes for a client.

    :ivar c2d_message_inbox: The C2D message Inbox.
    :ivar input_message_inboxes: A dictionary mapping input names to input message Inboxes.
    :ivar generic_method_request_inbox: The generic method request Inbox.
    :ivar named_method_request_inboxes: A dictionary mapping method names to method request Inboxes.
    """

    def __init__(self, inbox_type):
        """Initializer for the InboxManager.

        :param inbox_type: An Inbox class that the manager will use to create Inboxes.
        """
        self._create_inbox = inbox_type
        self.unified_message_inbox = self._create_inbox()
        self.generic_method_request_inbox = self._create_inbox()
        self.twin_patch_inbox = self._create_inbox()
        self.client_event_inbox = self._create_inbox()

        # These inboxes are used only for non-unified receives, using APIs which are now
        # deprecated on the client. However we need to keep them functional for backwards
        # compatibility
        self.c2d_message_inbox = self._create_inbox()
        self.input_message_inboxes = {}
        self.named_method_request_inboxes = {}

        # Set this value to True if want to only use unified message mode
        self.use_unified_msg_mode = False

    def get_unified_message_inbox(self):
        """Retrieve the Inbox for all messages (C2D and Input)"""
        return self.unified_message_inbox

    def get_input_message_inbox(self, input_name):
        """Retrieve the input message Inbox for a given input.

        If the Inbox does not already exist, it will be created.

        :param str input_name: The name of the input for which the associated Inbox is desired.
        :returns: An Inbox for input messages on the selected input.
        """
        try:
            inbox = self.input_message_inboxes[input_name]
        except KeyError:
            # Create new Inbox for input if it does not yet exist
            inbox = self._create_inbox()
            self.input_message_inboxes[input_name] = inbox

        return inbox

    def get_c2d_message_inbox(self):
        """Retrieve the Inbox for C2D messages.

        :returns: An Inbox for C2D messages.
        """
        return self.c2d_message_inbox

    def get_method_request_inbox(self, method_name=None):
        """Retrieve the method request Inbox for a given method name if provided,
        or for generic method requests if not.

        If the Inbox does not already exist, it will be created.

        :param str method_name: Optional. The name of the method for which the
        associated Inbox is desired.
        :returns: An Inbox for method requests.
        """
        if method_name:
            try:
                inbox = self.named_method_request_inboxes[method_name]
            except KeyError:
                # Create a new Inbox for the method name
                inbox = self._create_inbox()
                self.named_method_request_inboxes[method_name] = inbox
        else:
            inbox = self.generic_method_request_inbox

        return inbox

    def get_twin_patch_inbox(self):
        """Retrieve the Inbox for twin patches that arrive from the service

        :returns: An Inbox for twin patches
        """
        return self.twin_patch_inbox

    def get_client_event_inbox(self):
        """Retrieve the Inbox for events that occur within the client

        :returns: An Inbox for client events
        """
        return self.client_event_inbox

    def clear_all_method_requests(self):
        """Delete all method requests currently in inboxes."""
        self.generic_method_request_inbox.clear()
        for inbox in self.named_method_request_inboxes.values():
            inbox.clear()

    def route_input_message(self, incoming_message):
        """Route an incoming input message

        In unified message mode, route to the unified message inbox

        In standard mode, route to the corresponding input message Inbox. If the input
        is unknown, the message will be dropped.

        :param incoming_message: The message to be routed.

        :returns: Boolean indicating if message was successfully routed or not.
        """
        input_name = incoming_message.input_name
        if self.use_unified_msg_mode:
            # Put in the unified message inbox if in simplified mode
            self.unified_message_inbox.put(incoming_message)
            return True
        else:
            # If not in simplified mode, get a specific inbox for the input
            try:
                inbox = self.input_message_inboxes[input_name]
            except KeyError:
                logger.warning(
                    "No input message inbox for {} - dropping message".format(input_name)
                )
                return False
            else:
                inbox.put(incoming_message)
                logger.debug("Input message sent to {} inbox".format(input_name))
                return True

    def route_c2d_message(self, incoming_message):
        """Route an incoming C2D message

        In unified message mode, route to the unified message inbox.

        In standard mode, route to to the C2D message Inbox.

        :param incoming_message: The message to be routed.

        :returns: Boolean indicating if message was successfully routed or not.
        """
        if self.use_unified_msg_mode:
            # Put in the unified message inbox if in simplified mode
            self.unified_message_inbox.put(incoming_message)
            return True
        else:
            self.c2d_message_inbox.put(incoming_message)
            logger.debug("C2D message sent to inbox")
            return True

    def route_method_request(self, incoming_method_request):
        """Route an incoming method request to the correct method request Inbox.

        If the method name is recognized, it will be routed to a method-specific Inbox.
        Otherwise, it will be routed to the generic method request Inbox.

        :param incoming_method_request: The method request to be routed.

        :returns: Boolean indicating if the method request was successfully routed or not.
        """
        try:
            inbox = self.named_method_request_inboxes[incoming_method_request.name]
        except KeyError:
            inbox = self.generic_method_request_inbox
        inbox.put(incoming_method_request)
        return True

    def route_twin_patch(self, incoming_patch):
        """Route an incoming twin patch to the twin patch Inbox.

        :param incoming_patch: The patch to be routed.

        :returns: Boolean indicating if patch was successfully routed or not.
        """
        self.twin_patch_inbox.put(incoming_patch)
        logger.debug("twin patch message sent to inbox")
        return True



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
import json
import base64
import requests  # type: ignore
import requests_unixsocket
import urllib
from azure.iot.device.common.auth.signing_mechanism import SigningMechanism
from azure.iot.device import user_agent

requests_unixsocket.monkeypatch()
logger = logging.getLogger(__name__)


class IoTEdgeError(Exception):
    pass


class IoTEdgeHsm(SigningMechanism):
    """
    Constructor for instantiating a iot hsm object.  This is an object that
    communicates with the Azure IoT Edge HSM in order to get connection credentials
    for an Azure IoT Edge module.  The credentials that this object return come in
    two forms:

    1. The trust bundle, which is a certificate that can be used as a trusted cert
       to authenticate the SSL connection between the IoE Edge module and IoT Edge
    2. A signing function, which can be used to create the sig field for a
       SharedAccessSignature string which can be used to authenticate with Iot Edge
    """

    def __init__(self, module_id, generation_id, workload_uri, api_version):
        """
        Constructor for instantiating a Azure IoT Edge HSM object

        :param str module_id: The module id
        :param str api_version: The API version
        :param str generation_id: The module generation id
        :param str workload_uri: The workload uri
        """
        self.module_id = urllib.parse.quote(module_id, safe="")
        self.api_version = api_version
        self.generation_id = generation_id
        self.workload_uri = _format_socket_uri(workload_uri)

    def get_certificate(self):
        """
        Return the server verification certificate from the trust bundle that can be used to
        validate the server-side SSL TLS connection that we use to talk to Edge

        :return: The server verification certificate to use for connections to the Azure IoT Edge
        instance, as a PEM certificate in string form.

        :raises: IoTEdgeError if unable to retrieve the certificate.
        """
        r = requests.get(
            self.workload_uri + "trust-bundle",
            params={"api-version": self.api_version},
            headers={"User-Agent": urllib.parse.quote_plus(user_agent.get_iothub_user_agent())},
        )
        # Validate that the request was successful
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise IoTEdgeError("Unable to get trust bundle from Edge") from e
        # Decode the trust bundle
        try:
            bundle = r.json()
        except ValueError as e:
            raise IoTEdgeError("Unable to decode trust bundle") from e
        # Retrieve the certificate
        try:
            cert = bundle["certificate"]
        except KeyError as e:
            raise IoTEdgeError("No certificate in trust bundle") from e
        return cert

    def sign(self, data_str):
        """
        Use the IoTEdge HSM to sign a piece of string data.  The caller should then insert the
        returned value (the signature) into the 'sig' field of a SharedAccessSignature string.

        :param str data_str: The data string to sign

        :return: The signature, as a URI-encoded and base64-encoded value that is ready to
        directly insert into the SharedAccessSignature string.

        :raises: IoTEdgeError if unable to sign the data.
        """
        encoded_data_str = base64.b64encode(data_str.encode("utf-8")).decode()

        path = "{workload_uri}modules/{module_id}/genid/{gen_id}/sign".format(
            workload_uri=self.workload_uri, module_id=self.module_id, gen_id=self.generation_id
        )
        sign_request = {"keyId": "primary", "algo": "HMACSHA256", "data": encoded_data_str}

        r = requests.post(  # can we use json field instead of data?
            url=path,
            params={"api-version": self.api_version},
            headers={"User-Agent": urllib.parse.quote(user_agent.get_iothub_user_agent(), safe="")},
            data=json.dumps(sign_request),
        )
        try:
            r.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise IoTEdgeError("Unable to sign data") from e
        try:
            sign_response = r.json()
        except ValueError as e:
            raise IoTEdgeError("Unable to decode signed data") from e
        try:
            signed_data_str = sign_response["digest"]
        except KeyError as e:
            raise IoTEdgeError("No signed data received") from e

        return signed_data_str  # what format is this? string? bytes?


def _format_socket_uri(old_uri):
    """
    This function takes a socket URI in one form and converts it into another form.

    The source form is based on what we receive inside the IOTEDGE_WORKLOADURI
    environment variable, and it looks like this:
    "unix:///var/run/iotedge/workload.sock"

    The destination form is based on what the requests_unixsocket library expects
    and it looks like this:
    "http+unix://%2Fvar%2Frun%2Fiotedge%2Fworkload.sock/"

    The function changes the prefix, uri-encodes the path, and adds a slash
    at the end.

    If the socket URI does not start with unix:// this function only adds
    a slash at the end.

    :param old_uri: The URI in IOTEDGE_WORKLOADURI form

    :return: The URI in requests_unixsocket form
    """
    old_prefix = "unix://"
    new_prefix = "http+unix://"

    if old_uri.startswith(old_prefix):
        stripped_uri = old_uri[len(old_prefix) :]
        if stripped_uri.endswith("/"):
            stripped_uri = stripped_uri[:-1]
        new_uri = new_prefix + urllib.parse.quote(stripped_uri, safe="")
    else:
        new_uri = old_uri

    if not new_uri.endswith("/"):
        new_uri += "/"

    return new_uri



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
from azure.iot.device.common.pipeline.config import BasePipelineConfig

logger = logging.getLogger(__name__)


class IoTHubPipelineConfig(BasePipelineConfig):
    """A class for storing all configurations/options for IoTHub clients in the Azure IoT Python Device Client Library."""

    def __init__(
        self,
        hostname,
        device_id,
        module_id=None,
        product_info="",
        ensure_desired_properties=True,
        **kwargs
    ):
        """Initializer for IoTHubPipelineConfig which passes all unrecognized keyword-args down to BasePipelineConfig
        to be evaluated. This stacked options setting is to allow for unique configuration options to exist between the
        multiple clients, while maintaining a base configuration class with shared config options.

        :param str hostname: The hostname of the IoTHub to connect to
        :param str device_id: The device identity being used with the IoTHub
        :param str module_id: The module identity being used with the IoTHub
        :param str product_info: A custom identification string for the type of device connecting to Azure IoT Hub.
        :param bool ensure_desired_properties: Indicates if twin_patches should ensure the most
            recent desired properties patch has been received upon re-connections
        """
        super().__init__(hostname=hostname, **kwargs)

        # IoTHub Connection Details
        self.device_id = device_id
        self.module_id = module_id

        # Product Info
        self.product_info = product_info

        # Stage Behavior
        self.ensure_desired_properties = ensure_desired_properties

        # Now, the parameters below are not exposed to the user via kwargs. They need to be set by manipulating the IoTHubPipelineConfig object.
        # They are not in the BasePipelineConfig because these do not apply to the provisioning client.
        self.blob_upload = False
        self.method_invoke = False



"""Azure IoT Hub Device Library

This library provides functionality for communicating with the Azure IoT Hub
as a Device or Module.
"""

from .sync_clients import IoTHubDeviceClient, IoTHubModuleClient
from .models import Message, MethodRequest, MethodResponse

__all__ = ["IoTHubDeviceClient", "IoTHubModuleClient", "Message", "MethodRequest", "MethodResponse"]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
import threading
import abc
from azure.iot.device import constant


logger = logging.getLogger(__name__)

DEFAULT_KEEPALIVE = 60


class BasePipelineConfig(abc.ABC):
    """A base class for storing all configurations/options shared across the Azure IoT Python Device Client Library.
    More specific configurations such as those that only apply to the IoT Hub Client will be found in the respective
    config files.
    """

    def __init__(
        self,
        hostname,
        gateway_hostname=None,
        sastoken=None,
        x509=None,
        server_verification_cert=None,
        websockets=False,
        cipher="",
        proxy_options=None,
        keep_alive=DEFAULT_KEEPALIVE,
        auto_connect=True,
        connection_retry=True,
        connection_retry_interval=10,
    ):
        """Initializer for BasePipelineConfig

        :param str hostname: The hostname being connected to
        :param str gateway_hostname: The gateway hostname optionally being used
        :param sastoken: SasToken to be used for authentication. Mutually exclusive with x509.
        :type sastoken: :class:`azure.iot.device.common.auth.SasToken`
        :param x509: X509 to be used for authentication. Mutually exclusive with sastoken.
        :type x509: :class:`azure.iot.device.models.X509`
        :param str server_verification_cert: The trusted certificate chain.
            Necessary when using connecting to an endpoint which has a non-standard root of trust,
            such as a protocol gateway.
        :param bool websockets: Enabling/disabling websockets in MQTT. This feature is relevant
            if a firewall blocks port 8883 from use.
        :param cipher: Optional cipher suite(s) for TLS/SSL, as a string in
            "OpenSSL cipher list format" or as a list of cipher suite strings.
        :type cipher: str or list(str)
        :param proxy_options: Details of proxy configuration
        :type proxy_options: :class:`azure.iot.device.common.models.ProxyOptions`
        :param int keepalive: Maximum period in seconds between communications with the
            broker.
        :param bool auto_connect: Indicates if automatic connects should occur
        :param bool connection_retry: Indicates if dropped connection should result in attempts to
            re-establish it
        :param int connection_retry_interval: Interval (in seconds) between connection retries
        """
        # Network
        self.hostname = hostname
        self.gateway_hostname = gateway_hostname
        self.keep_alive = self._sanitize_keep_alive(keep_alive)

        # Auth
        self.sastoken = sastoken
        self.x509 = x509
        if (not sastoken and not x509) or (sastoken and x509):
            raise ValueError("One of either 'sastoken' or 'x509' must be provided")
        self.server_verification_cert = server_verification_cert
        self.websockets = websockets
        self.cipher = self._sanitize_cipher(cipher)
        self.proxy_options = proxy_options

        # Pipeline
        self.auto_connect = auto_connect
        self.connection_retry = connection_retry
        self.connection_retry_interval = self._sanitize_connection_retry_interval(
            connection_retry_interval
        )

    @staticmethod
    def _sanitize_cipher(cipher):
        """Sanitize the cipher input and convert to a string in OpenSSL list format"""
        if isinstance(cipher, list):
            cipher = ":".join(cipher)

        if isinstance(cipher, str):
            cipher = cipher.upper()
            cipher = cipher.replace("_", "-")
        else:
            raise TypeError("Invalid type for 'cipher'")

        return cipher

    @staticmethod
    def _sanitize_keep_alive(keep_alive):
        try:
            keep_alive = int(keep_alive)
        except (ValueError, TypeError):
            raise TypeError("Invalid type for 'keep alive'. Must be a numeric value.")

        if keep_alive <= 0:
            # Not allowing a keep alive of 0 as this would mean frequent ping exchanges.
            raise ValueError("'keep alive' must be greater than 0")

        if keep_alive > constant.MAX_KEEP_ALIVE_SECS:
            raise ValueError("'keep_alive' cannot exceed 1740 seconds (29 minutes)")

        return keep_alive

    @staticmethod
    def _sanitize_connection_retry_interval(connection_retry_interval):
        try:
            connection_retry_interval = int(connection_retry_interval)
        except (ValueError, TypeError):
            raise TypeError("Invalid type for 'connection_retry_interval'. Must be a numeric value")

        if connection_retry_interval > threading.TIMEOUT_MAX:
            # Python timers have a (platform dependent) max timeout.
            raise ValueError(
                "'connection_retry_interval' cannot exceed {} seconds".format(threading.TIMEOUT_MAX)
            )

        if connection_retry_interval <= 0:
            raise ValueError("'connection_retry_interval' must be greater than 0")

        return connection_retry_interval



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from azure.iot.device.common.pipeline import PipelineEvent


class C2DMessageEvent(PipelineEvent):
    """
    A PipelineEvent object which represents an incoming C2D event.  This object is probably
    created by some converter stage based on a protocol-specific event
    """

    def __init__(self, message):
        """
        Initializer for C2DMessageEvent objects.

        :param Message message: The Message object for the message that was received.
        """
        super().__init__()
        self.message = message


class InputMessageEvent(PipelineEvent):
    """
    A PipelineEvent object which represents an incoming input message event.  This object is probably
    created by some converter stage based on a protocol-specific event
    """

    def __init__(self, message):
        """
        Initializer for InputMessageEvent objects.

        :param Message message: The Message object for the message that was received. This message
            is expected to have had the .input_name attribute set
        """
        super().__init__()
        self.message = message


class MethodRequestEvent(PipelineEvent):
    """
    A PipelineEvent object which represents an incoming MethodRequest event.
    This object is probably created by some converter stage based on a protocol-specific event.
    """

    def __init__(self, method_request):
        super().__init__()
        self.method_request = method_request


class TwinDesiredPropertiesPatchEvent(PipelineEvent):
    """
    A PipelineEvent object which represents an incoming twin desired properties patch.  This
    object is probably created by some converter stage based on a protocol-specific event.
    """

    def __init__(self, patch):
        super().__init__()
        self.patch = patch



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
import traceback
import threading
import weakref
from . import (
    pipeline_ops_base,
    PipelineStage,
    pipeline_ops_mqtt,
    pipeline_events_mqtt,
    pipeline_thread,
    pipeline_exceptions,
    pipeline_events_base,
)
from azure.iot.device.common.mqtt_transport import MQTTTransport
from azure.iot.device.common import handle_exceptions, transport_exceptions

logger = logging.getLogger(__name__)

# Maximum amount of time we wait for ConnectOperation to complete
# TODO: This whole logic of timeout should probably be handled in the TimeoutStage
WATCHDOG_INTERVAL = 60


class MQTTTransportStage(PipelineStage):
    """
    PipelineStage object which is responsible for interfacing with the MQTT protocol wrapper object.
    This stage handles all MQTT operations and any other operations (such as ConnectOperation) which
    is not in the MQTT group of operations, but can only be run at the protocol level.
    """

    def __init__(self):
        super().__init__()

        # The transport will be instantiated upon receiving the InitializePipelineOperation
        self.transport = None
        # The current in-progress op that affects connection state (Connect, Disconnect, Reauthorize)
        self._pending_connection_op = None

    @pipeline_thread.runs_on_pipeline_thread
    def _cancel_pending_connection_op(self, error=None):
        """
        Cancel any running connect, disconnect or reauthorize connection op. Since our ability to "cancel" is fairly limited,
        all this does (for now) is to fail the operation
        """

        op = self._pending_connection_op
        if op:
            # NOTE: This code path should NOT execute in normal flow. There should never already be a pending
            # connection op when another is added, due to the ConnectionLock stage.
            # If this block does execute, there is a bug in the codebase.
            if not error:
                error = pipeline_exceptions.OperationCancelled(
                    "Cancelling because new ConnectOperation or DisconnectOperation was issued"
                )
            self._cancel_connection_watchdog(op)
            self._pending_connection_op = None
            op.complete(error=error)

    @pipeline_thread.runs_on_pipeline_thread
    def _start_connection_watchdog(self, connection_op):
        """
        Start a watchdog on the connection operation. This protects against cases where transport.connect()
        succeeds but the CONNACK never arrives. This is like a timeout, but it is handled at this level
        because specific cleanup needs to take place on timeout (see below), and this cleanup doesn't
        belong anywhere else since it is very specific to this stage.
        """
        logger.debug("{}({}): Starting watchdog".format(self.name, connection_op.name))

        self_weakref = weakref.ref(self)
        op_weakref = weakref.ref(connection_op)

        @pipeline_thread.invoke_on_pipeline_thread
        def watchdog_function():
            this = self_weakref()
            op = op_weakref()
            if this and op and this._pending_connection_op is op:
                logger.info(
                    "{}({}): Connection watchdog expired.  Cancelling op".format(this.name, op.name)
                )
                try:
                    this.transport.disconnect()
                except Exception:
                    # If we don't catch this, the pending connection op might not ever be cancelled.
                    # Most likely, the transport isn't actually connected, but other failures are theoretically
                    # possible. Either way, if disconnect fails, we should assume that we're disconnected.
                    logger.info(
                        "transport.disconnect raised error while disconnecting in watchdog.  Safe to ignore."
                    )
                    logger.info(traceback.format_exc())

                if this.nucleus.connected:

                    logger.info(
                        "{}({}): Pipeline is still connected on watchdog expiration.  Sending DisconnectedEvent".format(
                            this.name, op.name
                        )
                    )
                    this.send_event_up(pipeline_events_base.DisconnectedEvent())
                this._cancel_pending_connection_op(
                    error=pipeline_exceptions.OperationTimeout(
                        "Transport timeout on connection operation"
                    )
                )
            else:
                logger.debug("Connection watchdog expired, but pending op is not the same op")

        connection_op.watchdog_timer = threading.Timer(WATCHDOG_INTERVAL, watchdog_function)
        connection_op.watchdog_timer.daemon = True
        connection_op.watchdog_timer.start()

    @pipeline_thread.runs_on_pipeline_thread
    def _cancel_connection_watchdog(self, op):
        try:
            if op.watchdog_timer:
                logger.debug("{}({}): cancelling watchdog".format(self.name, op.name))
                op.watchdog_timer.cancel()
                op.watchdog_timer = None
        except AttributeError:
            pass

    @pipeline_thread.runs_on_pipeline_thread
    def _run_op(self, op):
        if isinstance(op, pipeline_ops_base.InitializePipelineOperation):

            # If there is a gateway hostname, use that as the hostname for connection,
            # rather than the hostname itself
            if self.nucleus.pipeline_configuration.gateway_hostname:
                logger.debug(
                    "Gateway Hostname Present. Setting Hostname to: {}".format(
                        self.nucleus.pipeline_configuration.gateway_hostname
                    )
                )
                hostname = self.nucleus.pipeline_configuration.gateway_hostname
            else:
                logger.debug(
                    "Gateway Hostname not present. Setting Hostname to: {}".format(
                        self.nucleus.pipeline_configuration.hostname
                    )
                )
                hostname = self.nucleus.pipeline_configuration.hostname

            # Create the Transport object, set it's handlers
            logger.debug("{}({}): got connection args".format(self.name, op.name))
            self.transport = MQTTTransport(
                client_id=op.client_id,
                hostname=hostname,
                username=op.username,
                server_verification_cert=self.nucleus.pipeline_configuration.server_verification_cert,
                x509_cert=self.nucleus.pipeline_configuration.x509,
                websockets=self.nucleus.pipeline_configuration.websockets,
                cipher=self.nucleus.pipeline_configuration.cipher,
                proxy_options=self.nucleus.pipeline_configuration.proxy_options,
                keep_alive=self.nucleus.pipeline_configuration.keep_alive,
            )
            self.transport.on_mqtt_connected_handler = self._on_mqtt_connected
            self.transport.on_mqtt_connection_failure_handler = self._on_mqtt_connection_failure
            self.transport.on_mqtt_disconnected_handler = self._on_mqtt_disconnected
            self.transport.on_mqtt_message_received_handler = self._on_mqtt_message_received

            # There can only be one pending connection operation (Connect, Disconnect)
            # at a time. The existing one must be completed or canceled before a new one is set.

            # Currently, this means that if, say, a connect operation is the pending op and is executed
            # but another connection op is begins by the time the CONNACK is received, the original
            # operation will be cancelled, but the CONNACK for it will still be received, and complete the
            # NEW operation. This is not desirable, but it is how things currently work.

            # We are however, checking the type, so the CONNACK from a cancelled Connect, cannot successfully
            # complete a Disconnect operation.

            # Note that a ReauthorizeConnectionOperation will never be pending because it will
            # instead spawn separate Connect and Disconnect operations.
            self._pending_connection_op = None

            op.complete()

        elif isinstance(op, pipeline_ops_base.ShutdownPipelineOperation):
            try:
                self.transport.shutdown()
            except Exception as e:
                logger.info("transport.shutdown raised error")
                logger.info(traceback.format_exc())
                op.complete(error=e)
            else:
                op.complete()

        elif isinstance(op, pipeline_ops_base.ConnectOperation):
            logger.debug("{}({}): connecting".format(self.name, op.name))

            self._cancel_pending_connection_op()
            self._pending_connection_op = op
            self._start_connection_watchdog(op)
            # Use SasToken as password if present. If not present (e.g. using X509),
            # then no password is required because auth is handled via other means.
            if self.nucleus.pipeline_configuration.sastoken:
                password = str(self.nucleus.pipeline_configuration.sastoken)
            else:
                password = None
            try:
                self.transport.connect(password=password)
            except Exception as e:
                logger.info("transport.connect raised error")
                logger.info(traceback.format_exc())
                self._cancel_connection_watchdog(op)
                self._pending_connection_op = None
                op.complete(error=e)

        elif isinstance(op, pipeline_ops_base.DisconnectOperation):
            logger.debug("{}({}): disconnecting".format(self.name, op.name))

            self._cancel_pending_connection_op()
            self._pending_connection_op = op
            # We don't need a watchdog on disconnect because there's no callback to wait for
            # and we respond to a watchdog timeout by calling disconnect, which is what we're
            # already doing.

            try:
                # The connect after the disconnect will be triggered upon completion of the
                # disconnect in the on_disconnected handler
                self.transport.disconnect(clear_inflight=op.hard)
            except Exception as e:
                logger.info("transport.disconnect raised error while disconnecting")
                logger.info(traceback.format_exc())
                self._pending_connection_op = None
                op.complete(error=e)

        elif isinstance(op, pipeline_ops_base.ReauthorizeConnectionOperation):
            logger.debug(
                "{}({}): reauthorizing. Will issue disconnect and then a connect".format(
                    self.name, op.name
                )
            )
            self_weakref = weakref.ref(self)
            reauth_op = op  # rename for clarity

            def on_disconnect_complete(op, error):
                this = self_weakref()
                if error:
                    # Failing a disconnect should still get us disconnected, so can proceed anyway
                    logger.debug(
                        "Disconnect failed during reauthorization, continuing with connect"
                    )
                connect_op = reauth_op.spawn_worker_op(pipeline_ops_base.ConnectOperation)

                # NOTE: this relies on the fact that before the disconnect is completed it is
                # unset as the pending connection op. Otherwise there would be issues here.
                this.run_op(connect_op)

            disconnect_op = pipeline_ops_base.DisconnectOperation(callback=on_disconnect_complete)
            disconnect_op.hard = False

            self.run_op(disconnect_op)

        elif isinstance(op, pipeline_ops_mqtt.MQTTPublishOperation):
            logger.debug("{}({}): publishing on {}".format(self.name, op.name, op.topic))

            @pipeline_thread.invoke_on_pipeline_thread_nowait
            def on_complete(cancelled=False):
                if cancelled:
                    op.complete(
                        error=pipeline_exceptions.OperationCancelled(
                            "Operation cancelled before PUBACK received"
                        )
                    )
                else:
                    logger.debug(
                        "{}({}): PUBACK received. completing op.".format(self.name, op.name)
                    )
                    op.complete()

            try:
                self.transport.publish(topic=op.topic, payload=op.payload, callback=on_complete)
            except Exception as e:
                op.complete(error=e)

        elif isinstance(op, pipeline_ops_mqtt.MQTTSubscribeOperation):
            logger.debug("{}({}): subscribing to {}".format(self.name, op.name, op.topic))

            @pipeline_thread.invoke_on_pipeline_thread_nowait
            def on_complete(cancelled=False):
                if cancelled:
                    op.complete(
                        error=pipeline_exceptions.OperationCancelled(
                            "Operation cancelled before SUBACK received"
                        )
                    )
                else:
                    logger.debug(
                        "{}({}): SUBACK received. completing op.".format(self.name, op.name)
                    )
                    op.complete()

            try:
                self.transport.subscribe(topic=op.topic, callback=on_complete)
            except Exception as e:
                op.complete(error=e)

        elif isinstance(op, pipeline_ops_mqtt.MQTTUnsubscribeOperation):
            logger.debug("{}({}): unsubscribing from {}".format(self.name, op.name, op.topic))

            @pipeline_thread.invoke_on_pipeline_thread_nowait
            def on_complete(cancelled=False):
                if cancelled:
                    op.complete(
                        error=pipeline_exceptions.OperationCancelled(
                            "Operation cancelled before UNSUBACK received"
                        )
                    )
                else:
                    logger.debug(
                        "{}({}): UNSUBACK received.  completing op.".format(self.name, op.name)
                    )
                    op.complete()

            try:
                self.transport.unsubscribe(topic=op.topic, callback=on_complete)
            except Exception as e:
                op.complete(error=e)

        else:
            # This code block should not be reached in correct program flow.
            # This will raise an error when executed.
            self.send_op_down(op)

    @pipeline_thread.invoke_on_pipeline_thread_nowait
    def _on_mqtt_message_received(self, topic, payload):
        """
        Handler that gets called by the protocol library when an incoming message arrives.
        Convert that message into a pipeline event and pass it up for someone to handle.
        """
        logger.debug("{}: message received on topic {}".format(self.name, topic))
        self.send_event_up(
            pipeline_events_mqtt.IncomingMQTTMessageEvent(topic=topic, payload=payload)
        )

    @pipeline_thread.invoke_on_pipeline_thread_nowait
    def _on_mqtt_connected(self):
        """
        Handler that gets called by the transport when it connects.
        """
        logger.info("_on_mqtt_connected called")
        # Send an event to tell other pipeline stages that we're connected. Do this before
        # we do anything else (in case upper stages have any "are we connected" logic.
        self.send_event_up(pipeline_events_base.ConnectedEvent())

        if isinstance(self._pending_connection_op, pipeline_ops_base.ConnectOperation):
            logger.debug("{}: completing connect op".format(self.name))
            op = self._pending_connection_op
            self._cancel_connection_watchdog(op)
            self._pending_connection_op = None
            op.complete()
        else:
            # This should indicate something odd is going on.
            # If this occurs, either a connect was completed while there was no pending op,
            # OR that a connect was completed while a disconnect op was pending
            logger.info(
                "{}: Connection was unexpected (no connection op pending)".format(self.name)
            )

    @pipeline_thread.invoke_on_pipeline_thread_nowait
    def _on_mqtt_connection_failure(self, cause):
        """
        Handler that gets called by the transport when a connection fails.

        :param Exception cause: The Exception that caused the connection failure.
        """

        logger.info("{}: _on_mqtt_connection_failure called: {}".format(self.name, cause))

        if isinstance(self._pending_connection_op, pipeline_ops_base.ConnectOperation):
            logger.debug("{}: failing connect op".format(self.name))
            op = self._pending_connection_op
            self._cancel_connection_watchdog(op)
            self._pending_connection_op = None
            op.complete(error=cause)
        else:
            logger.debug("{}: Connection failure was unexpected".format(self.name))
            handle_exceptions.swallow_unraised_exception(
                cause,
                log_msg="Unexpected connection failure (no pending operation). Safe to ignore.",
                log_lvl="info",
            )

    @pipeline_thread.invoke_on_pipeline_thread_nowait
    def _on_mqtt_disconnected(self, cause=None):
        """
        Handler that gets called by the transport when the transport disconnects.

        :param Exception cause: The Exception that caused the disconnection, if any (optional)
        """
        if cause:
            logger.info("{}: _on_mqtt_disconnect called: {}".format(self.name, cause))
        else:
            logger.info("{}: _on_mqtt_disconnect called".format(self.name))

        # Send an event to tell other pipeline stages that we're disconnected. Do this before
        # we do anything else (in case upper stages have any "are we connected" logic.)
        # NOTE: Other stages rely on the fact that this occurs before any op that may be in
        # progress is completed. Be careful with changing the order things occur here.
        self.send_event_up(pipeline_events_base.DisconnectedEvent())

        if self._pending_connection_op:

            op = self._pending_connection_op

            if isinstance(op, pipeline_ops_base.DisconnectOperation):
                logger.debug(
                    "{}: Expected disconnect - completing pending disconnect op".format(self.name)
                )
                # Swallow any errors if we intended to disconnect - even if something went wrong, we
                # got to the state we wanted to be in!
                if cause:
                    handle_exceptions.swallow_unraised_exception(
                        cause,
                        log_msg="Unexpected error while disconnecting - swallowing error",
                    )
                # Disconnect complete, no longer pending
                self._pending_connection_op = None
                op.complete()

            else:
                logger.debug(
                    "{}: Unexpected disconnect - completing pending {} operation".format(
                        self.name, op.name
                    )
                )
                # Cancel any potential connection watchdog, and clear the pending op
                self._cancel_connection_watchdog(op)
                self._pending_connection_op = None
                # Complete
                if cause:
                    op.complete(error=cause)
                else:
                    op.complete(
                        error=transport_exceptions.ConnectionDroppedError("transport disconnected")
                    )
        else:
            logger.info("{}: Unexpected disconnect (no pending connection op)".format(self.name))

            # If there is no connection retry, cancel any transport operations waiting on response
            # so that they do not get stuck there.
            if not self.nucleus.pipeline_configuration.connection_retry:
                logger.debug(
                    "{}: Connection Retry disabled - cancelling in-flight operations".format(
                        self.name
                    )
                )
                # TODO: Remove private access to the op manager (this layer shouldn't know about it)
                # This is a stopgap. I didn't want to invest too much infrastructure into a cancel flow
                # given that future development of individual operation cancels might affect the
                # approach to cancelling inflight ops waiting in the transport.
                self.transport._op_manager.cancel_all_operations()

            # Regardless of cause, it is now a ConnectionDroppedError. Log it and swallow it.
            # Higher layers will see that we're disconnected and may reconnect as necessary.
            e = transport_exceptions.ConnectionDroppedError("Unexpected disconnection")
            e.__cause__ = cause
            self.report_background_exception(e)



"""Azure Provisioning Device Communication Pipeline

This package provides pipeline for use with the Azure Provisioning Device SDK.

INTERNAL USAGE ONLY
"""
from .mqtt_pipeline import MQTTPipeline  # noqa: F401
from .config import ProvisioningPipelineConfig  # noqa: F401



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import logging
from . import pipeline_exceptions
from . import pipeline_thread

logger = logging.getLogger(__name__)


class PipelineOperation(object):
    """
    A base class for data objects representing operations that travels down the pipeline.

    Each PipelineOperation object represents a single asynchronous operation that is performed
    by the pipeline. The PipelineOperation objects travel through "stages" of the pipeline,
    and each stage has the opportunity to act on each specific operation that it
    receives.  If a stage does not handle a particular operation, it needs to pass it to the
    next stage.  If the operation gets to the end of the pipeline without being handled
    (completed), then it is treated as an error.

    :ivar name: The name of the operation.  This is used primarily for logging
    :type name: str
    :ivar callback: The callback that is called when the operation is completed, either
        successfully or with a failure.
    :type callback: Function
    :ivar needs_connection: This is an attribute that indicates whether a particular operation
        requires a connection to operate.  This is currently used by the AutoConnectStage
        stage, but this functionality will be revamped shortly.
    :type needs_connection: Boolean
    :ivar error: The presence of a value in the error attribute indicates that the operation failed,
        absence of this value indicates that the operation either succeeded or hasn't been handled yet.
    :type error: Error
    """

    def __init__(self, callback):
        """
        Initializer for PipelineOperation objects.

        :param Function callback: The function that gets called when this operation is complete or has
            failed. The callback function must accept A PipelineOperation object which indicates
            the specific operation which has completed or failed.
        """
        if self.__class__ == PipelineOperation:
            raise TypeError(
                "Cannot instantiate PipelineOperation object.  You need to use a derived class"
            )
        self.name = self.__class__.__name__
        self.callback_stack = []
        self.needs_connection = False
        self.completed = False  # Operation has been fully completed
        self.completing = False  # Operation is in the process of completing
        self.error = None  # Error associated with Operation completion

        self.add_callback(callback)

    def add_callback(self, callback):
        """Adds a callback to the Operation that will be triggered upon Operation completion.

        When an Operation is completed, all callbacks will be resolved in LIFO order.

        Callbacks cannot be added to an already completed operation, or an operation that is
        currently undergoing a completion process.

        :param callback: The callback to add to the operation.

        :raises: OperationError if the operation is already completed, or is in the process of
            completing.
        """
        if self.completed:
            raise pipeline_exceptions.OperationError(
                "{}: Attempting to add a callback to an already-completed operation!".format(
                    self.name
                )
            )
        if self.completing:
            raise pipeline_exceptions.OperationError(
                "{}: Attempting to add a callback to a operation with completion in progress!".format(
                    self.name
                )
            )
        else:
            self.callback_stack.append(callback)

    @pipeline_thread.runs_on_pipeline_thread
    def complete(self, error=None):
        """Complete the operation, and trigger all callbacks in LIFO order.

        The operation is completed successfully be default, or completed unsuccessfully if an error
        is provided.

        An operation that is already fully completed, or in the process of completion cannot be
        completed again.

        This process can be halted if a callback for the operation invokes the .halt_completion()
        method on this Operation.

        Note that if an error is raised the operation may not be able to be completed.

        :param error: Optionally provide an Exception object indicating the error that caused
            the completion. Providing an error indicates that the operation was unsuccessful.

        :raises: OperationError if the operation cannot properly reach a completed state
        :raises: OperationError if an error occurs in resolving any callbacks
        """
        if error:
            logger.debug("{}: completing with error {}".format(self.name, error))
        else:
            logger.debug("{}: completing without error".format(self.name))

        if self.completed or self.completing:
            raise pipeline_exceptions.OperationError(
                "Attempting to complete an already-completed operation: {}".format(self.name)
            )
        else:
            # Operation is now in the process of completing
            self.completing = True
            self.error = error

            while self.callback_stack:
                if not self.completing:
                    logger.debug("{}: Completion halted!".format(self.name))
                    break
                if self.completed:
                    # This block should never be reached - this is an invalid state.
                    # If this block is reached, there is a bug in the code.
                    self.halt_completion()
                    raise pipeline_exceptions.OperationError(
                        "Operation reached fully completed state while still resolving completion: {}".format(
                            self.name
                        )
                    )

                callback = self.callback_stack.pop()
                try:
                    callback(op=self, error=error)
                except Exception as e:
                    logger.warning(
                        "Unhandled error while triggering callback for {}".format(self.name)
                    )
                    self.halt_completion()
                    raise pipeline_exceptions.OperationError(
                        "Exception occurred while triggering completion callback"
                    ) from e

            if self.completing:
                # Operation is now completed, no longer in the process of completing
                self.completing = False
                self.completed = True

    @pipeline_thread.runs_on_pipeline_thread
    def halt_completion(self):
        """Halt the completion of an operation that is currently undergoing a completion process
        as a result of a call to .complete().

        Completion cannot be halted if there is no currently ongoing completion process. The only
        way to successfully invoke this method is from within a callback on the Operation in
        question.

        This method will leave any yet-untriggered callbacks on the Operation to be triggered upon
        a later completion.

        This method will clear any error associated with the currently ongoing completion process
        from the Operation.
        """
        if not self.completing:
            raise pipeline_exceptions.OperationError(
                "Attempting to halt completion of an operation not in the process of completion: {}".format(
                    self.name
                )
            )
        else:
            self.completing = False
            self.error = None
            logger.debug("{}: Operation completion halted".format(self.name))

    @pipeline_thread.runs_on_pipeline_thread
    def spawn_worker_op(self, worker_op_type, **kwargs):
        """Create and return a new operation, which, when completed, will complete the operation
        it was spawned from.

        :param worker_op_type: The type (class) of the new worker operation.
        :param **kwargs: The arguments to instantiate the new worker operation with. Note that a
            callback is not required, but if provided, will be triggered prior to completing the
            operation that spawned the worker operation.

        :returns: A new worker operation of the type specified in the worker_op_type parameter.
        """
        logger.debug("{}: creating worker op of type {}".format(self.name, worker_op_type.__name__))

        @pipeline_thread.runs_on_pipeline_thread
        def on_worker_op_complete(op, error):
            logger.debug("{}: Worker op ({}) has been completed".format(self.name, op.name))
            self.complete(error=error)

        if "callback" in kwargs:
            provided_callback = kwargs["callback"]
            kwargs["callback"] = on_worker_op_complete
            worker_op = worker_op_type(**kwargs)
            worker_op.add_callback(provided_callback)
        else:
            kwargs["callback"] = on_worker_op_complete
            worker_op = worker_op_type(**kwargs)

        return worker_op


class InitializePipelineOperation(PipelineOperation):
    """
    A PipelineOperation for doing initial setup of the pipeline

    Attributes can be dynamically added to this operation for use in other stages if necessary
    (e.g. initialization requires a derived value)
    """

    pass


class ShutdownPipelineOperation(PipelineOperation):
    """
    A PipelineOperation for doing teardown of the pipeline.
    """

    pass


class ConnectOperation(PipelineOperation):
    """
    A PipelineOperation object which tells the pipeline to connect to whatever service it needs to connect to.

    This operation is in the group of base operations because connecting is a common operation that many clients might need to do.

    Even though this is an base operation, it will most likely be handled by a more specific stage (such as an IoTHub or MQTT stage).
    """

    def __init__(self, callback):
        self.watchdog_timer = None
        super().__init__(callback)


class ReauthorizeConnectionOperation(PipelineOperation):
    """
    A PipelineOperation object which tells the pipeline to reauthorize the connection to whatever service it is connected to.

    Clients will most-likely submit a ReauthorizeConnectionOperation when some credential (such as a sas token) has changed and the protocol client
    needs to re-establish the connection to refresh the credentials

    This operation is in the group of base operations because reauthorizing is a common operation that many clients might need to do.

    Even though this is an base operation, it will most likely be handled by a more specific stage (such as an IoTHub or MQTT stage).
    """

    pass


class DisconnectOperation(PipelineOperation):
    """
    A PipelineOperation object which tells the pipeline to disconnect from whatever service it might be connected to.

    This operation is in the group of base operations because disconnecting is a common operation that many clients might need to do.

    Even though this is an base operation, it will most likely be handled by a more specific stage (such as an IoTHub or MQTT stage).
    """

    def __init__(self, callback):
        self.hard = True  # Indicates if this is a "hard" disconnect that kills in-flight ops
        super().__init__(callback)


class EnableFeatureOperation(PipelineOperation):
    """
    A PipelineOperation object which tells the pipeline to "enable" a particular feature.

    A "feature" is just a string which represents some set of functionality that needs to be enabled, such as "C2D" or "Twin".

    This object has no notion of what it means to "enable" a feature.  That knowledge is handled by stages in the pipeline which might convert
    this operation to a more specific operation (such as an MQTT subscribe operation with a specific topic name).

    This operation is in the group of base operations because disconnecting is a common operation that many clients might need to do.

    Even though this is an base operation, it will most likely be handled by a more specific stage (such as an IoTHub or MQTT stage).
    """

    def __init__(self, feature_name, callback):
        """
        Initializer for EnableFeatureOperation objects.

        :param str feature_name: Name of the feature that is being enabled.  The meaning of this
            string is defined in the stage which handles this operation.
        :param Function callback: The function that gets called when this operation is complete or has
            failed.  The callback function must accept A PipelineOperation object which indicates
            the specific operation which has completed or failed.
        """
        super().__init__(callback=callback)
        self.feature_name = feature_name


class DisableFeatureOperation(PipelineOperation):
    """
    A PipelineOperation object which tells the pipeline to "disable" a particular feature.

    A "feature" is just a string which represents some set of functionality that needs to be disabled, such as "C2D" or "Twin".

    This object has no notion of what it means to "disable" a feature.  That knowledge is handled by stages in the pipeline which might convert
    this operation to a more specific operation (such as an MQTT unsubscribe operation with a specific topic name).

    This operation is in the group of base operations because disconnecting is a common operation that many clients might need to do.

    Even though this is an base operation, it will most likely be handled by a more specific stage (such as an IoTHub or MQTT stage).
    """

    def __init__(self, feature_name, callback):
        """
        Initializer for DisableFeatureOperation objects.

        :param str feature_name: Name of the feature that is being disabled.  The meaning of this
            string is defined in the stage which handles this operation.
        :param Function callback: The function that gets called when this operation is complete or has
            failed.  The callback function must accept A PipelineOperation object which indicates
            the specific operation which has completed or failed.
        """
        super().__init__(callback=callback)
        self.feature_name = feature_name


class RequestAndResponseOperation(PipelineOperation):
    """
    A PipelineOperation object which wraps the common operation of sending a request to iothub with a request_id ($rid)
    value and waiting for a response with the same $rid value.  This convention is used by both Twin and Provisioning
    features.

    Even though this is an base operation, it will most likely be generated and also handled by more specifics stages
    (such as IoTHub or MQTT stages).

    The type of the request payload and the response payload is undefined at this level.  The type of the payload is defined
    based on the type of request that is being executed.  If types need to be converted, that is the responsibility of
    the stage which creates this operation, and also the stage which executes on the operation.

    :ivar status_code: The status code returned by the response.  Any value under 300 is considered success.
    :type status_code: int
    :ivar response_body: The body of the response.
    :type response_body: Undefined
    :ivar query_params: Any query parameters that need to be sent with the request.
    Example is the id of the operation as returned by the initial provisioning request.
    """

    def __init__(
        self, request_type, method, resource_location, request_body, callback, query_params=None
    ):
        """
        Initializer for RequestAndResponseOperation objects

        :param str request_type: The type of request.  This is a string which is used by protocol-specific stages to
            generate the actual request.  For example, if request_type is "twin", then the iothub_mqtt stage will convert
            the request into an MQTT publish with topic that begins with $iothub/twin
        :param str method: The method for the request, in the REST sense of the word, such as "POST", "GET", etc.
        :param str resource_location: The resource that the method is acting on, in the REST sense of the word.
            For twin request with method "GET", this is most likely the string "/" which retrieves the entire twin
        :param request_body: The body of the request.  This is a required field, and a single space can be used to denote
            an empty body.
        :type request_body: Undefined
        :param Function callback: The function that gets called when this operation is complete or has
            failed.  The callback function must accept A PipelineOperation object which indicates
            the specific operation which has completed or failed.
        """
        super().__init__(callback=callback)
        self.request_type = request_type
        self.method = method
        self.resource_location = resource_location
        self.request_body = request_body
        self.status_code = None
        self.response_body = None
        self.query_params = query_params


class RequestOperation(PipelineOperation):
    """
    A PipelineOperation object which is the first part of an RequestAndResponseOperation operation (the request). The second
    part of the RequestAndResponseOperation operation (the response) is returned via an ResponseEvent event.

    Even though this is an base operation, it will most likely be generated and also handled by more specifics stages
    (such as IoTHub or MQTT stages).
    """

    def __init__(
        self,
        request_type,
        method,
        resource_location,
        request_body,
        request_id,
        callback,
        query_params=None,
    ):
        """
        Initializer for RequestOperation objects

        :param str request_type: The type of request.  This is a string which is used by protocol-specific stages to
            generate the actual request.  For example, if request_type is "twin", then the iothub_mqtt stage will convert
            the request into an MQTT publish with topic that begins with $iothub/twin
        :param str method: The method for the request, in the REST sense of the word, such as "POST", "GET", etc.
        :param str resource_location: The resource that the method is acting on, in the REST sense of the word.
            For twin request with method "GET", this is most likely the string "/" which retrieves the entire twin
        :param request_body: The body of the request.  This is a required field, and a single space can be used to denote
            an empty body.
        :type request_body: dict, str, int, float, bool, or None (JSON compatible values)
        :param Function callback: The function that gets called when this operation is complete or has
            failed.  The callback function must accept A PipelineOperation object which indicates
            the specific operation which has completed or failed.
        :type query_params: Any query parameters that need to be sent with the request.
        Example is the id of the operation as returned by the initial provisioning request.
        """
        super().__init__(callback=callback)
        self.method = method
        self.resource_location = resource_location
        self.request_type = request_type
        self.request_body = request_body
        self.request_id = request_id
        self.query_params = query_params
